from dotenv import load_dotenv

load_dotenv()

from datetime import datetime
import json
import time
from unittest.mock import patch
from freezegun import freeze_time

from langchain.chains import LLMMathChain
from langchain.llms import OpenAI
from langchain.chat_models import ChatOpenAI
from langchain.prompts import PromptTemplate, ChatPromptTemplate
from langchain.schema import (
    BaseOutputParser,
)
from langchain.chains import LLMChain
from langwatch.types import TraceMetadata
import openai
from pytest_httpx import HTTPXMock
import requests_mock

import langwatch
import langwatch.langchain
from langchain.agents import load_tools, initialize_agent, Tool
from langchain.agents.agent import AgentExecutor
from langchain.agents.openai_functions_agent.base import OpenAIFunctionsAgent

from tests.utils import *  # type: ignore


class CommaSeparatedListOutputParser(BaseOutputParser):
    """Parse the output of an LLM call to a comma-separated list."""

    def parse(self, text: str):
        """Parse the output of an LLM call."""
        return text.strip().split(", ")


class TestLangChainTracer:
    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_lang_chain_chat_calls(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_response(
            json=create_openai_chat_completion_mock("red, blue, green, yellow"),
            url="https://api.openai.com/v1/chat/completions",
        )

        template = """You are a helpful assistant who generates comma separated lists.
        A user will pass in a category, and you should generate 5 objects in that category in a comma separated list.
        ONLY return a comma separated list, and nothing more."""
        human_template = "{text}"

        chat_prompt = ChatPromptTemplate.from_messages(
            [
                ("system", template),
                ("human", human_template),
            ]
        )
        chain = LLMChain(
            llm=ChatOpenAI(),
            prompt=chat_prompt,
            output_parser=CommaSeparatedListOutputParser(),
        )
        with langwatch.langchain.LangChainTracer(
            metadata={
                "user_id": "user-123",
                "thread_id": "thread-456",
                "customer_id": "customer-789",
            }
        ) as langWatchCallback:
            result = chain.run(text="colors", callbacks=[langWatchCallback])
        assert result == ["red", "blue", "green", "yellow"]

        time.sleep(0.01)
        request_history = [
            r for r in requests_mock.request_history if langwatch.endpoint in r.url
        ]
        trace_request = request_history[0].json()
        assert trace_request["metadata"]["user_id"] == "user-123"
        assert trace_request["metadata"]["thread_id"] == "thread-456"
        assert trace_request["metadata"]["customer_id"] == "customer-789"

        second_span, first_span = trace_request["spans"]

        assert first_span["type"] == "chain"
        assert first_span["trace_id"]
        assert first_span["parent_id"]
        assert first_span["timestamps"]["started_at"]
        assert first_span["timestamps"]["finished_at"]
        assert first_span["output"] == {
            "type": "json",
            "value": {"text": ["red", "blue", "green", "yellow"]},
        }

        assert second_span["type"] == "llm"
        assert second_span["trace_id"]
        assert second_span["span_id"].startswith("span_")
        assert second_span["parent_id"] == first_span["span_id"]
        assert second_span["model"] == "openai/gpt-3.5-turbo"
        assert second_span["input"] == {
            "type": "chat_messages",
            "value": [
                {"role": "system", "content": template},
                {"role": "user", "content": "colors"},
            ],
        }
        assert second_span["output"] == {
            "type": "chat_messages",
            "value": [
                {"role": "assistant", "content": "red, blue, green, yellow"},
            ],
        }

        assert second_span["params"] == {"n": 1, "temperature": 0.7}
        assert second_span["metrics"] == {
            "prompt_tokens": 5,
            "completion_tokens": 16,
        }
        assert second_span["timestamps"]["started_at"]
        assert (
            second_span["timestamps"]["finished_at"]
            > second_span["timestamps"]["started_at"]
        )

    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_lang_chain_completion(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_response(
            json=create_openai_completion_mock("red, blue, green, yellow"),
            url="https://api.openai.com/v1/completions",
        )

        template = """You are a helpful assistant who generates comma separated lists.
        A user will pass in a category, and you should generate 5 objects in that category in a comma separated list.
        ONLY return a comma separated list, and nothing more. Make a list of {text}"""

        chat_prompt = PromptTemplate.from_template(template)
        chain = LLMChain(
            llm=OpenAI(),
            prompt=chat_prompt,
        )
        with langwatch.langchain.LangChainTracer() as langWatchCallback:
            result = chain.run(text="colors", callbacks=[langWatchCallback])
        assert result == "red, blue, green, yellow"

        time.sleep(0.01)
        request_history = [
            r for r in requests_mock.request_history if langwatch.endpoint in r.url
        ]
        first_span, second_span = request_history[0].json()["spans"]

        assert first_span["type"] == "chain"

        assert second_span["type"] == "llm"
        assert second_span["vendor"] == "openai"
        assert second_span["model"] == "gpt-3.5-turbo-instruct"
        assert second_span["input"] == {
            "type": "json",
            "value": [template.replace("{text}", "colors")],
        }
        assert second_span["outputs"] == [
            {
                "type": "text",
                "value": "red, blue, green, yellow",
            }
        ]
        assert second_span["params"] == {"temperature": 0.7, "stream": False}
        assert second_span["metrics"] == {
            "prompt_tokens": 5,
            "completion_tokens": 16,
        }
        assert second_span["timestamps"]["started_at"] >= int(
            datetime(2022, 1, 1, 0, 0, 0).timestamp() * 1000
        )
        assert (
            second_span["timestamps"]["finished_at"]
            > second_span["timestamps"]["started_at"]
        )

    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_lang_chain_streaming_chat_calls(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_response(
            stream=create_openai_chat_completion_stream_mock(
                ["red,", " blue,", " green,", " yellow"]
            ),
            url="https://api.openai.com/v1/chat/completions",
        )

        template = """You are a helpful assistant who generates comma separated lists.
        A user will pass in a category, and you should generate 5 objects in that category in a comma separated list.
        ONLY return a comma separated list, and nothing more."""
        human_template = "{text}"

        chat_prompt = ChatPromptTemplate.from_messages(
            [
                ("system", template),
                ("human", human_template),
            ]
        )
        chain = LLMChain(
            llm=ChatOpenAI(streaming=True),
            prompt=chat_prompt,
            # TODO: output_parser=CommaSeparatedListOutputParser(),
        )
        with langwatch.langchain.LangChainTracer() as langWatchCallback:
            result = chain.run(text="colors", callbacks=[langWatchCallback])
        assert result == "red, blue, green, yellow"

        time.sleep(0.01)
        request_history = [
            r for r in requests_mock.request_history if langwatch.endpoint in r.url
        ]
        first_span, second_span = request_history[0].json()["spans"]

        assert first_span["type"] == "chain"
        assert first_span["trace_id"].startswith("trace_")
        assert first_span["parent_id"] == None
        assert first_span["timestamps"]["started_at"] >= int(
            datetime(2022, 1, 1, 0, 0, 0).timestamp() * 1000
        )
        assert (
            first_span["timestamps"]["finished_at"]
            >= first_span["timestamps"]["started_at"]
        )

        assert second_span["type"] == "llm"
        assert second_span["trace_id"].startswith("trace_")
        assert second_span["span_id"].startswith("span_")
        assert second_span["parent_id"] == first_span["span_id"]
        assert second_span["vendor"] == "openai"
        assert second_span["model"] == "gpt-3.5-turbo"
        assert second_span["input"] == {
            "type": "chat_messages",
            "value": [
                {"role": "system", "content": template},
                {"role": "user", "content": "colors"},
            ],
        }
        assert second_span["outputs"] == [
            {
                "type": "chat_messages",
                "value": [
                    {"role": "assistant", "content": "red, blue, green, yellow"},
                ],
            }
        ]
        assert second_span["params"] == {"temperature": 0.7, "stream": True}
        assert "metrics" not in second_span
        assert second_span["timestamps"]["started_at"] >= int(
            datetime(2022, 1, 1, 0, 0, 0).timestamp() * 1000
        )
        assert (
            second_span["timestamps"]["finished_at"]
            > second_span["timestamps"]["started_at"]
        )

    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_agents_functions_and_tools(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):  # TODO: test without functions?
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_callback(
            one_mock_at_a_time(
                [
                    {
                        "choices": [
                            {
                                "finish_reason": "function_call",
                                "index": 0,
                                "message": {
                                    "content": None,
                                    "function_call": {
                                        "arguments": '{\n  "__arg1": "2+2"\n}',
                                        "name": "Calculator",
                                    },
                                    "role": "assistant",
                                },
                            }
                        ],
                        "created": 1697189473,
                        "id": "chatcmpl-898nhOctWsU6TpiJYEsrwcAHj1sSN",
                        "model": "gpt-3.5-turbo-0613",
                        "object": "chat.completion",
                        "usage": {
                            "completion_tokens": 19,
                            "prompt_tokens": 66,
                            "total_tokens": 85,
                        },
                    },
                    create_openai_chat_completion_mock(
                        '```text\n2 + 2\n```\n...numexpr.evaluate("2 + 2")...\n'
                    ),
                    create_openai_chat_completion_mock("2 + 2 is equal to 4."),
                ],
            ),
            url="https://api.openai.com/v1/chat/completions",
        )

        llm = ChatOpenAI()
        llm_math_chain = LLMMathChain.from_llm(llm=llm, verbose=True)
        tools = [
            Tool(
                name="Calculator",
                func=llm_math_chain.run,
                description="useful for when you need to answer questions about math",
            ),
        ]
        agent = AgentExecutor.from_agent_and_tools(
            agent=OpenAIFunctionsAgent.from_llm_and_tools(
                llm,
                tools,
                verbose=True,
            ),
            tools=tools,
            verbose=True,
        )

        with langwatch.langchain.LangChainTracer() as langWatchCallback:
            result = agent.run("how much is 2+2?", callbacks=[langWatchCallback])

        assert result == "2 + 2 is equal to 4."

        time.sleep(0.01)
        request_history = [
            r for r in requests_mock.request_history if langwatch.endpoint in r.url
        ]
        spans = request_history[0].json()["spans"]

        first_llm_call = [span for span in spans if span["type"] == "llm"][0]

        assert first_llm_call["params"]["functions"] == [
            {
                "name": "Calculator",
                "description": "useful for when you need to answer questions about math",
                "parameters": {
                    "properties": {"__arg1": {"title": "__arg1", "type": "string"}},
                    "required": ["__arg1"],
                    "type": "object",
                },
            }
        ]

        calculator_tool = [
            span
            for span in spans
            if span["type"] == "tool"
            and "name" in span
            and span["name"] == "Calculator"
        ][0]
        assert calculator_tool["outputs"] == [{"type": "text", "value": "Answer: 4"}]

        calculator_agent = [span for span in spans if span["type"] == "agent"][0]
        assert calculator_agent["input"] == {
            "type": "json",
            "value": {"input": "how much is 2+2?"},
        }
        assert calculator_agent["outputs"] == [
            {"type": "json", "value": {"output": "2 + 2 is equal to 4."}}
        ]

    def test_trace_errors(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_exception(
            Exception("An error occurred!"),
            url="https://api.openai.com/v1/chat/completions",
        )

        chain = LLMChain(
            llm=ChatOpenAI(max_retries=0),
            prompt=ChatPromptTemplate.from_messages(
                [
                    ("human", "hi there"),
                ]
            ),
        )
        with langwatch.langchain.LangChainTracer() as langWatchCallback:
            try:
                chain.run(text="hi", callbacks=[langWatchCallback])
            except:
                pass

        time.sleep(0.01)
        request_history = [
            r for r in requests_mock.request_history if langwatch.endpoint in r.url
        ]
        first_span, second_span = request_history[0].json()["spans"]

        assert first_span["error"]["message"] != None
        assert second_span["error"]["message"] != None

    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_lang_chain_expression_language(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_response(
            json=create_openai_chat_completion_mock(
                "Why don't bears wear shoes?\n\nBecause they have bear feet!"
            ),
            url="https://api.openai.com/v1/chat/completions",
        )

        model = ChatOpenAI()
        prompt = ChatPromptTemplate.from_template("tell me a joke about {topic}")
        chain = prompt | model

        with langwatch.langchain.LangChainTracer() as langWatchCallback:
            result = chain.invoke(
                {"topic": "bears"}, config={"callbacks": [langWatchCallback]}
            )
        assert (
            result.content
            == "Why don't bears wear shoes?\n\nBecause they have bear feet!"
        )

        time.sleep(0.501)
        request_history = [
            r for r in requests_mock.request_history if langwatch.endpoint in r.url
        ]
        first_span = request_history[0].json()["spans"][0]

        assert first_span["type"] == "chain"
        assert first_span["outputs"] == [
            {
                "type": "chat_messages",
                "value": [
                    {
                        "role": "assistant",
                        "content": "Why don't bears wear shoes?\n\nBecause they have bear feet!",
                    }
                ],
            }
        ]

    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_adding_metadata_in_the_middle_of_the_process(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_response(
            json=create_openai_chat_completion_mock(
                "Why don't bears wear shoes?\n\nBecause they have bear feet!"
            ),
            url="https://api.openai.com/v1/chat/completions",
        )

        model = ChatOpenAI()
        prompt = ChatPromptTemplate.from_template("tell me a joke about {topic}")
        chain = prompt | model

        with langwatch.langchain.LangChainTracer(
            metadata={"labels": []}
        ) as langWatchCallback:
            result = chain.invoke(
                {"topic": "bears"}, config={"callbacks": [langWatchCallback]}
            )
            langWatchCallback.metadata["labels"] += ["joke"]  # type: ignore
        assert (
            result.content
            == "Why don't bears wear shoes?\n\nBecause they have bear feet!"
        )

        time.sleep(0.501)
        request_history = [
            r for r in requests_mock.request_history if langwatch.endpoint in r.url
        ]
        first_span = request_history[0].json()["spans"][0]
        metadata = request_history[0].json()["metadata"]

        assert first_span["type"] == "chain"
        assert first_span["outputs"] == [
            {
                "type": "chat_messages",
                "value": [
                    {
                        "role": "assistant",
                        "content": "Why don't bears wear shoes?\n\nBecause they have bear feet!",
                    }
                ],
            }
        ]
        assert metadata["labels"] == ["joke"]



import json
from typing import Any, List
import httpx
from pytest_httpx import IteratorStream


def one_mock_at_a_time(mocks: List[Any]):
    mock_index = 0

    def _one_mock_at_a_time(_request: httpx.Request):
        nonlocal mock_index
        response = httpx.Response(
            status_code=200,
            json=mocks[mock_index],
        )
        mock_index += 1
        return response

    return _one_mock_at_a_time


def create_openai_completion_mock(text):
    return {
        "id": "cmpl-861BvC6rh12Y1hrk8ml1BaQPJP0mE",
        "object": "text_completion",
        "created": 1696445239,
        "model": "gpt-3.5-turbo-instruct",
        "choices": [
            {
                "text": text,
                "index": 0,
                "logprobs": None,
                "finish_reason": "length",
            }
        ],
        "system_fingerprint": None,
        "usage": {"prompt_tokens": 5, "completion_tokens": 16, "total_tokens": 21},
    }


def create_openai_completion_stream_mock(*text_groups):
    return IteratorStream(
        [
            str.encode(
                "data: "
                + json.dumps(create_openai_completion_chunk(index, text))
                + "\n\n"
            )
            for index, texts in enumerate(text_groups)
            for text in texts
        ]
        + [str.encode("data: [DONE]")]
    )


async def create_openai_completion_async_stream_mock(*text_groups):
    for index, texts in enumerate(text_groups):
        for text in texts:
            yield create_openai_completion_chunk(index, text)


def create_openai_completion_chunk(index: int, text: str):
    return {
        "id": "cmpl-86MtN6iz0JSSIiLAesNKKqyDtKcZO",
        "object": "text_completion",
        "created": 1696528657,
        "choices": [
            {"text": text, "index": index, "logprobs": None, "finish_reason": "length"}
        ],
        "model": "gpt-3.5-turbo-instruct",
        "system_fingerprint": None,
        "usage": None,
    }


def create_openai_chat_completion_mock(text):
    return {
        "id": "chatcmpl-86zIvz53Wa4qTc1ksUt3coF5yTvm7",
        "object": "chat.completion",
        "created": 1696676313,
        "model": "gpt-3.5-turbo-0613",
        "choices": [
            {
                "index": 0,
                "logprobs": None,
                "message": {
                    "role": "assistant",
                    "content": text,
                    "function_call": None,
                    "tool_calls": None,
                },
                "finish_reason": "stop",
            }
        ],
        "usage": {"prompt_tokens": 5, "completion_tokens": 16, "total_tokens": 21},
        "system_fingerprint": None,
    }


def create_openai_chat_completion_stream_mock(*text_groups):
    list = []
    for index in range(0, len(text_groups)):
        list.append(
            create_openai_chat_completion_chunk(
                index, {"role": "assistant", "content": ""}
            )
        )
    for index, texts in enumerate(text_groups):
        for text in texts:
            list.append(create_openai_chat_completion_chunk(index, {"content": text}))
    for index in range(0, len(text_groups)):
        list.append(create_openai_chat_completion_chunk(index, {}))

    return IteratorStream(
        [str.encode("data: " + json.dumps(item) + "\n\n") for item in list]
        + [str.encode("data: [DONE]")]
    )


def create_openai_chat_completion_function_stream_mock(*function_calls):
    list = []
    for index, function_call in enumerate(function_calls):
        list.append(
            create_openai_chat_completion_chunk(
                index,
                {
                    "role": "assistant",
                    "content": None,
                    "function_call": {"name": function_call["name"], "arguments": ""},
                },
            )
        )
    for index, function_call in enumerate(function_calls):
        for token_index, token in enumerate(function_call["arguments"].split(" ")):
            list.append(
                create_openai_chat_completion_chunk(
                    index,
                    {
                        "function_call": {
                            "arguments": ("" if token_index == 0 else " ") + token
                        }
                    },
                )
            )
    for index in range(0, len(function_calls)):
        list.append(create_openai_chat_completion_chunk(index, {}))

    return IteratorStream(
        [str.encode("data: " + json.dumps(item) + "\n\n") for item in list]
        + [str.encode("data: [DONE]")]
    )


def create_openai_chat_completion_chunk(index: int, delta: dict):
    empty_delta = {
        "content": None,
        "function_call": None,
        "tool_calls": None,
        "role": None,
    }
    return {
        "id": "chatcmpl-871IjS1cTmejs3MVHu3zJseODF5KU",
        "object": "chat.completion.chunk",
        "created": 1696683989,
        "model": "gpt-3.5-turbo-0613",
        "choices": [
            {"index": index, "delta": empty_delta | delta, "finish_reason": None}
        ],
        "system_fingerprint": None,
    }



import time
from dotenv import load_dotenv

load_dotenv()

import chainlit as cl

import langwatch


@cl.on_message
async def main(message: cl.Message):
    msg = cl.Message(
        content="",
    )

    with langwatch.trace() as trace:
        with trace.span(
            type="llm",
            input=message.content,
        ) as span:
            time.sleep(1)  # generating the message...
            generated_message = "Hello there! How can I help from context manager?"

            span.update(output=generated_message)

        await msg.stream_token(generated_message)

    await msg.update()



from contextlib import contextmanager
from typing import List, Literal, Optional, Union, cast
from uuid import UUID

import httpx
from pydantic import BaseModel, ConfigDict

import langwatch
import langwatch.tracer
from langwatch.tracer import ContextSpan, ContextTrace, get_current_trace
from langwatch.types import (
    Conversation,
    EvaluationResult,
    Money,
    SpanMetrics,
    RAGChunk,
    SpanTypes,
    TypedValueEvaluationResult,
    TypedValueGuardrailResult,
    TypedValueJson,
)


class EvaluationResultModel(BaseModel):
    status: Literal["processed", "skipped", "error"]
    passed: Optional[bool] = None
    score: Optional[float] = None
    details: Optional[str] = None
    label: Optional[str] = None
    cost: Optional[Money] = None
    error_type: Optional[str] = None


def evaluate(
    slug: str,
    name: Optional[str] = None,
    input: Optional[str] = None,
    output: Optional[str] = None,
    expected_output: Optional[str] = None,
    contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
    expected_contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
    conversation: Optional[Conversation] = None,
    settings: Optional[dict] = None,
    as_guardrail: bool = False,
    trace: Optional[ContextTrace] = None,
    span: Optional[ContextSpan] = None,
    api_key: Optional[str] = None,
):
    with _optional_create_span(
        trace=trace,
        span=span,
        name=name or slug,
        type="guardrail" if as_guardrail else "evaluation",
    ) as span:
        request_params = prepare_data(
            slug=slug,
            name=name,
            input=input,
            output=output,
            expected_output=expected_output,
            contexts=contexts,
            expected_contexts=expected_contexts,
            conversation=conversation,
            settings=settings,
            trace_id=span.trace.trace_id if span and span.trace else None,
            span_id=span.span_id if span else None,
            span=span,
            as_guardrail=as_guardrail,
            api_key=api_key,
        )
        try:
            with httpx.Client(timeout=900) as client:
                response = client.post(**request_params)
                response.raise_for_status()
        except Exception as e:
            return handle_exception(e, span, as_guardrail)

        return handle_response(response.json(), span, as_guardrail)


async def async_evaluate(
    slug: str,
    name: Optional[str] = None,
    input: Optional[str] = None,
    output: Optional[str] = None,
    expected_output: Optional[str] = None,
    contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
    expected_contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
    conversation: Optional[Conversation] = None,
    settings: Optional[dict] = None,
    as_guardrail: bool = False,
    trace: Optional[ContextTrace] = None,
    span: Optional[ContextSpan] = None,
    api_key: Optional[str] = None,
):
    with _optional_create_span(
        trace=trace,
        span=span,
        name=name or slug,
        type="guardrail" if as_guardrail else "evaluation",
    ) as span:
        request_params = prepare_data(
            slug=slug,
            name=name,
            input=input,
            output=output,
            expected_output=expected_output,
            contexts=contexts,
            expected_contexts=expected_contexts,
            conversation=conversation,
            settings=settings,
            trace_id=span.trace.trace_id if span and span.trace else None,
            span_id=span.span_id if span else None,
            span=span,
            as_guardrail=as_guardrail,
            api_key=api_key,
        )
        try:
            async with httpx.AsyncClient(timeout=900) as client:
                response = await client.post(**request_params)
                response.raise_for_status()
        except Exception as e:
            return handle_exception(e, span, as_guardrail)

        return handle_response(response.json(), span, as_guardrail)


@contextmanager
def _optional_create_span(
    trace: Optional[ContextTrace],
    span: Optional[ContextSpan],
    name: str,
    type: SpanTypes,
):
    trace_or_span = None
    if span:
        trace_or_span = span
    elif trace:
        trace_or_span = trace
    else:
        try:
            trace_or_span = get_current_trace()
        except:
            pass

    if trace_or_span:
        with trace_or_span.span(name=name, type=type) as span:
            yield span
    else:
        yield None


def prepare_data(
    slug: str,
    name: Optional[str],
    input: Optional[str],
    output: Optional[str],
    expected_output: Optional[str],
    contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
    expected_contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
    conversation: Optional[Conversation] = None,
    settings: Optional[dict] = None,
    trace_id: Optional[Union[str, UUID]] = None,
    span_id: Optional[Union[str, UUID]] = None,
    span: Optional[ContextSpan] = None,
    as_guardrail: bool = False,
    api_key: Optional[str] = None,
):
    data = {}
    if input is not None:
        data["input"] = input
    if output is not None:
        data["output"] = output
    if expected_output is not None:
        data["expected_output"] = expected_output
    if contexts is not None:
        data["contexts"] = contexts
    if expected_contexts is not None:
        data["expected_contexts"] = expected_contexts
    if conversation is not None:
        data["conversation"] = conversation
    if span:
        span.update(
            input=TypedValueJson(type="json", value=data),
            params=settings,  # type: ignore
        )

    return {
        "url": langwatch.endpoint + f"/api/evaluations/{slug}/evaluate",
        "json": {
            "trace_id": str(trace_id) if trace_id else None,
            "span_id": str(span_id) if span_id else None,
            "name": name,
            "data": data,
            "settings": settings,
            "as_guardrail": as_guardrail,
        },
        "headers": {"X-Auth-Token": str(api_key or langwatch.api_key)},
    }


def handle_response(
    response: dict,
    span: Optional[ContextSpan] = None,
    as_guardrail: bool = False,
):
    result = EvaluationResultModel.model_validate(response)
    if span:
        span.update(
            output=(
                TypedValueGuardrailResult(
                    type="guardrail_result",
                    value=cast(
                        EvaluationResult,
                        result.model_dump(exclude_unset=True, exclude_none=True),
                    ),
                )
                if as_guardrail
                else TypedValueEvaluationResult(
                    type="evaluation_result",
                    value=cast(
                        EvaluationResult,
                        result.model_dump(exclude_unset=True, exclude_none=True),
                    ),
                )
            )
        )
        if result.cost:
            span.update(
                metrics=SpanMetrics(
                    cost=result.cost.amount,
                )
            )
    return result


def handle_exception(
    e: Exception, span: Optional[ContextSpan] = None, as_guardrail: bool = False
):
    response: dict = {
        "status": "error",
        "details": repr(e),
    }
    if as_guardrail:
        response["passed"] = True
    return handle_response(
        response,
        span,
        as_guardrail,
    )



from contextlib import contextmanager
import random
import re
import time
import warnings
import dspy
from typing import Callable, List, Optional, Any, Type, Union
from typing_extensions import TypedDict
import langwatch
import httpx
import json
from pydantic import BaseModel
from dspy.predict import Predict
from dspy.teleprompt import (
    Teleprompter,
    BootstrapFewShot,
    BootstrapFewShotWithRandomSearch,
    COPRO,
    MIPRO,
    MIPROv2,
)
from dspy.signatures.signature import SignatureMeta
from dspy.primitives.prediction import Prediction, Completions
from dspy.primitives.example import Example
from pydantic.fields import FieldInfo
from coolname import generate_slug
from retry import retry
from dspy.evaluate.evaluate import Evaluate

from langwatch.tracer import ContextTrace
from langwatch.utils import reduce_payload_size
import litellm


class SerializableAndPydanticEncoder(json.JSONEncoder):
    def default(self, o):
        classname = f"{o.__class__.__module__}.{o.__class__.__name__}"
        if isinstance(o, FieldInfo):
            return {"__class__": classname} | {
                "field_type": o.json_schema_extra.get("__dspy_field_type", None),  # type: ignore
                "prefix": o.json_schema_extra.get("prefix", None),  # type: ignore
                "desc": o.json_schema_extra.get("desc", None),  # type: ignore
            }
        if isinstance(o, set):
            return list(o)
        if isinstance(o, Prediction):
            return {"__class__": classname} | {
                **o.__dict__.get("_store", {}),
                **{
                    k: v
                    for k, v in o.__dict__.items()
                    if k not in ["_store", "_completions"]
                },
            }
        if isinstance(o, Predict):
            return {"__class__": classname} | o.__dict__
        if isinstance(o, Example):
            return {"__class__": classname} | o.__dict__
        if isinstance(o, SignatureMeta):
            return {"__class__": classname} | {
                "signature": o.signature,
                "instructions": o.instructions,
                "fields": o.fields,
            }
        if isinstance(o, dspy.LM):
            return {"__class__": classname} | {
                "model": o.model,
                "kwargs": {
                    k: v
                    for k, v in o.kwargs.items()
                    if k in ["temperature", "max_tokens"]
                },
            }
        if isinstance(o, Completions):
            return {"__class__": classname} | o.__dict__
        if isinstance(o, BaseModel):
            return o.model_dump(exclude_unset=True)
        try:
            return super().default(o)
        except:
            return str(o)


class DSPyLLMCall(TypedDict, total=False):
    __class__: str
    model: Optional[str] = None
    response: Any
    prompt_tokens: Optional[int] = None
    completion_tokens: Optional[int] = None
    cost: Optional[float] = None


class Timestamps(BaseModel):
    created_at: int


class DSPyTrace(BaseModel):
    input: Any
    pred: Any


class DSPyExample(BaseModel):
    example: Any
    pred: Any
    score: float
    trace: Optional[List[DSPyTrace]]


class DSPyPredictor(BaseModel):
    name: str
    predictor: Any


class DSPyOptimizer(BaseModel):
    name: str
    parameters: Any


class DSPyStep(BaseModel):
    run_id: str
    experiment_slug: str
    workflow_version_id: Optional[str] = None
    index: str
    score: float
    label: str
    optimizer: DSPyOptimizer
    predictors: List[DSPyPredictor]
    examples: List[DSPyExample]
    llm_calls: List[DSPyLLMCall]
    timestamps: Timestamps


class LangWatchDSPy:
    """LangWatch DSPy visualization tracker"""

    _instance: Optional["LangWatchDSPy"] = None
    experiment_slug: Optional[str] = None
    experiment_path: str = ""
    run_id: Optional[str] = None

    examples_buffer: List[DSPyExample] = []
    llm_calls_buffer: List[DSPyLLMCall] = []
    steps_buffer: List[DSPyStep] = []

    def __new__(cls):
        """
        Singleton Pattern. See https://python-patterns.guide/gang-of-four/singleton/
        """

        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def init(
        self,
        experiment: str,
        optimizer: Optional[Teleprompter],
        run_id: Optional[str] = None,
        slug: Optional[str] = None,
        workflow_id: Optional[str] = None,
        workflow_version_id: Optional[str] = None,
    ):
        if langwatch.api_key is None:
            print("API key was not detected, calling langwatch.login()...")
            langwatch.login()
            return

        try:
            response = httpx.post(
                f"{langwatch.endpoint}/api/experiment/init",
                headers={"X-Auth-Token": langwatch.api_key or ""},
                json={
                    "experiment_slug": slug or experiment,
                    "experiment_type": "DSPY",
                    "experiment_name": experiment,
                    "workflow_id": workflow_id,
                },
                timeout=60,
            )
        except Exception as e:
            raise Exception(f"Error initializing LangWatch experiment: {e}")
        if response.status_code == 401:
            langwatch.api_key = None
            raise ValueError(
                "API key is not valid, please try to login again with langwatch.login()"
            )
        response.raise_for_status()

        self.experiment_slug = slug or experiment
        random.seed()  # MIPRO meses up the global seed, so we need to reset it to random to get a new run_id
        self.run_id = run_id or generate_slug(3)
        self.workflow_version_id = workflow_version_id
        self.reset()

        self.patch_llms()
        if optimizer is not None:
            self.patch_optimizer(optimizer)
        else:
            print(
                "No optimizer provided, assuming custom optimizer tracking, make sure to call `track_metric` and `log_step` manually: https://docs.langwatch.ai/dspy-visualization/custom-optimizer"
            )

        result = response.json()
        self.experiment_path = result["path"]
        print(f"\n[LangWatch] Experiment initialized, run_id: {self.run_id}")
        print(
            f"[LangWatch] Open {langwatch.endpoint}{self.experiment_path}?runIds={self.run_id} to track your DSPy training session live\n"
        )

    def patch_optimizer(self, optimizer: Teleprompter):
        METRIC_TRACKING_CLASSMAP = {
            BootstrapFewShot: LangWatchTrackedBootstrapFewShot,
            BootstrapFewShotWithRandomSearch: LangWatchTrackedBootstrapFewShotWithRandomSearch,
            COPRO: LangWatchTrackedCOPRO,
            MIPRO: LangWatchTrackedMIPRO,
            MIPROv2: LangWatchTrackedMIPROv2,
        }

        if optimizer.__class__ in METRIC_TRACKING_CLASSMAP:
            optimizer.__class__ = METRIC_TRACKING_CLASSMAP[optimizer.__class__]
            optimizer.patch()  # type: ignore
        else:
            supported = ", ".join(
                [f"{c.__name__}" for c in METRIC_TRACKING_CLASSMAP.keys()]
            )
            raise ValueError(
                f"Optimizer {optimizer.__class__.__name__} is not supported by LangWatch DSPy visualizer yet, only [{supported}] are supported, please open an issue: https://github.com/langwatch/langwatch/issues"
            )

    def reset(self):
        self.examples_buffer = []
        self.llm_calls_buffer = []
        self.steps_buffer = []

    def patch_llms(self):
        if not hasattr(dspy.OpenAI, "_original_request"):
            dspy.OpenAI._original_request = dspy.OpenAI.request  # type: ignore

        if not hasattr(dspy.AzureOpenAI, "_original_request"):
            dspy.AzureOpenAI._original_request = dspy.AzureOpenAI.request  # type: ignore

        if not hasattr(dspy.LM, "_original_call"):
            dspy.LM._original_call = dspy.LM.__call__  # type: ignore

        this = self

        def patched_request(self, *args, **kwargs):
            classname = f"{self.__class__.__module__}.{self.__class__.__name__}"
            response = self._original_request(*args, **kwargs)
            this.llm_calls_buffer.append(
                DSPyLLMCall(__class__=classname, response=response)
            )
            return response

        def patched_call(self, *args, **kwargs):
            classname = f"{self.__class__.__module__}.{self.__class__.__name__}"
            outputs = self._original_call(*args, **kwargs)
            if len(self.history) > 0:
                entry = self.history[-1]
                lm_response = entry["response"]

                llm_call = DSPyLLMCall(
                    __class__=classname,
                    model=self.model,
                )

                response = {}
                if len(outputs) == 1:
                    response["output"] = outputs[0]
                else:
                    response["outputs"] = outputs
                if "prompt" in entry:
                    response["prompt"] = entry["prompt"]
                if "messages" in entry:
                    response["messages"] = entry["messages"]
                if "model" in lm_response:
                    response["model"] = lm_response["model"]
                if "choices" in lm_response:
                    response["choices"] = lm_response["choices"]
                if (
                    not "_hidden_params" in lm_response
                    or "additional_headers" not in lm_response["_hidden_params"]
                ):
                    response["cached"] = True
                llm_call["response"] = response

                if "usage" in entry:
                    if "prompt_tokens" in entry["usage"]:
                        llm_call["prompt_tokens"] = entry["usage"]["prompt_tokens"]
                    if "completion_tokens" in entry["usage"]:
                        llm_call["completion_tokens"] = entry["usage"][
                            "completion_tokens"
                        ]

                if "cost" in entry:
                    llm_call["cost"] = entry["cost"]

                this.llm_calls_buffer.append(llm_call)
            return outputs

        dspy.OpenAI.request = patched_request
        dspy.AzureOpenAI.request = patched_request
        dspy.LM.__call__ = patched_call

    def track_metric(
        self,
        metric_fn: Callable[
            [Example, Prediction, Optional[List[Any]]], Union[bool, float]
        ],
    ):
        def wrapped(example, pred, trace=None):
            score = metric_fn(example, pred, trace=trace)  # type: ignore

            self.examples_buffer.append(
                DSPyExample(
                    example=example._store if hasattr(example, "_store") else example,
                    pred=pred._store if hasattr(pred, "_store") else pred,
                    score=float(score),
                    trace=(
                        [DSPyTrace(input=t[1], pred=t[2]) for t in trace]
                        if trace
                        else None
                    ),
                )
            )

            return score

        return wrapped

    def log_step(
        self,
        *,
        optimizer: DSPyOptimizer,
        index: str,
        score: float,
        label: str,
        predictors: List[DSPyPredictor],
    ):
        step = DSPyStep(
            run_id=self.run_id or "unknown",
            experiment_slug=self.experiment_slug or "unknown",
            index=index,
            score=score,
            label=label,
            optimizer=optimizer,
            predictors=predictors,
            examples=self.examples_buffer,
            llm_calls=self.llm_calls_buffer,
            workflow_version_id=self.workflow_version_id,
            timestamps=Timestamps(created_at=int(time.time() * 1000)),
        )
        self.steps_buffer.append(step)
        self.examples_buffer = []
        self.llm_calls_buffer = []
        self.send_steps()

    @retry(tries=3, delay=0.5)
    def send_steps(self):
        data_list = json.loads(
            json.dumps(self.steps_buffer, cls=SerializableAndPydanticEncoder)
        )
        data = [
            reduce_payload_size(
                item,
                max_string_length=5000,
                max_list_dict_length=-1,
            )
            for item in data_list
        ]
        response = httpx.post(
            f"{langwatch.endpoint}/api/dspy/log_steps",
            headers={
                "X-Auth-Token": langwatch.api_key or "",
                "Content-Type": "application/json",
            },
            data=json.dumps(data),  # type: ignore
            timeout=60,
        )
        response.raise_for_status()
        self.steps_buffer = []

    def tracer(self, trace: ContextTrace):
        return DSPyTracer(trace=trace)


langwatch_dspy = LangWatchDSPy()


class LangWatchTrackedBootstrapFewShot(BootstrapFewShot):
    last_step: int = 0
    last_round_idx: int = 0

    def patch(self):
        self.metric = langwatch_dspy.track_metric(self.metric)

    def _bootstrap_one_example(self, example, round_idx=0):
        self.last_round_idx = round_idx
        if round_idx != self.last_step and round_idx < self.max_rounds:
            self._log_step(round_idx)
        return super()._bootstrap_one_example(example, round_idx=round_idx)

    def _train(self):
        result = super()._train()
        final_predictors = [
            DSPyPredictor(name=name, predictor=predictor)
            for name, predictor in self.student.named_predictors()
        ]
        self._log_step(round_idx=self.last_round_idx + 1, predictors=final_predictors)
        return result

    def _log_step(self, round_idx: int, predictors: List[DSPyPredictor] = []):
        self.last_step = round_idx

        bootstrapped_demos_count = sum(
            [len(demos) for demos in self.name2traces.values()]
        )

        if len(predictors) == 0:
            student_ = self.student.deepcopy()
            for name, predictor in student_.named_predictors():
                augmented_demos = self.name2traces[name][: self.max_bootstrapped_demos]
                predictor.demos = augmented_demos
                predictors.append(DSPyPredictor(name=name, predictor=predictor))

        langwatch_dspy.log_step(
            optimizer=DSPyOptimizer(
                name=BootstrapFewShot.__name__,
                parameters={
                    "max_bootstrapped_demos": self.max_bootstrapped_demos,
                    "max_labeled_demos": self.max_labeled_demos,
                    "max_rounds": self.max_rounds,
                    "max_errors": self.max_errors,
                    "metric_threshold": self.metric_threshold,
                },
            ),
            index=str(round_idx),
            score=bootstrapped_demos_count,
            label="bootstrapped demos",
            predictors=predictors,
        )


class LangWatchTrackedBootstrapFewShotWithRandomSearch(
    BootstrapFewShotWithRandomSearch
):
    def patch(self):
        self.metric = langwatch_dspy.track_metric(self.metric)

    def compile(self, student, **kwargs):
        with self._patch_evaluate():
            return super().compile(student, **kwargs)

    @contextmanager
    def _patch_evaluate(self):
        original_evaluate_call = Evaluate.__call__
        step = 1

        this = self

        def patched_evaluate_call(self, program: dspy.Module, *args, **kwargs):
            nonlocal step

            if "return_all_scores" not in kwargs or not kwargs["return_all_scores"]:
                raise ValueError(
                    "return_all_scores is not True for some reason, please report it at https://github.com/langwatch/langwatch/issues"
                )
            score, subscores = original_evaluate_call(self, program, *args, **kwargs)  # type: ignore

            langwatch_dspy.log_step(
                optimizer=DSPyOptimizer(
                    name=BootstrapFewShotWithRandomSearch.__name__,
                    parameters={
                        "max_num_samples": this.max_num_samples,
                        "max_labeled_demos": this.max_labeled_demos,
                        "max_rounds": this.max_rounds,
                        "num_candidate_sets": this.num_candidate_sets,
                        "num_threads": this.num_threads,
                        "max_errors": this.max_errors,
                        "stop_at_score": this.stop_at_score,
                        "metric_threshold": this.metric_threshold,
                    },
                ),
                index=str(step),
                score=score,
                label="score",
                predictors=[
                    DSPyPredictor(name=name, predictor=predictor)
                    for name, predictor in program.named_predictors()
                ],
            )

            step += 1

            return score, subscores

        Evaluate.__call__ = patched_evaluate_call

        try:
            yield
        finally:
            Evaluate.__call__ = original_evaluate_call


class LangWatchTrackedCOPRO(COPRO):
    def patch(self):
        self.metric = langwatch_dspy.track_metric(self.metric)

    def compile(self, student, **kwargs):
        with self._patch_logger_and_evaluate():
            return super().compile(student, **kwargs)

    @contextmanager
    def _patch_logger_and_evaluate(self):
        original_logger_info = dspy.logger.info
        original_evaluate_call = Evaluate.__call__
        step = None
        scores = []

        this = self

        def patched_logger_info(text, *args, **kwargs):
            nonlocal step, scores

            match = re.search(
                r"At Depth (\d+)/(\d+), Evaluating Prompt Candidate #(\d+)/(\d+) for Predictor (\d+) of (\d+)",
                text,
            )
            if match:
                depth, _, breadth, _, predictor, _ = match.groups()
                new_step = f"{depth}.{predictor}.{breadth}"
                if new_step != step:
                    step = new_step
                    scores = []
            return original_logger_info(text, *args, **kwargs)

        def patched_evaluate_call(self, program: dspy.Module, *args, **kwargs):
            nonlocal step
            if not step:
                raise ValueError(
                    "Step is not defined, please report it at https://github.com/langwatch/langwatch/issues"
                )

            step_ = step

            score: float = original_evaluate_call(self, program, *args, **kwargs)  # type: ignore

            scores.append(score)

            if max(scores) == score:
                langwatch_dspy.log_step(
                    optimizer=DSPyOptimizer(
                        name=COPRO.__name__,
                        parameters={
                            "breadth": this.breadth,
                            "depth": this.depth,
                            "init_temperature": this.init_temperature,
                        },
                    ),
                    index=step_,
                    score=score,
                    label="score",
                    predictors=[
                        DSPyPredictor(name=name, predictor=predictor)
                        for name, predictor in program.named_predictors()
                    ],
                )

            return score

        dspy.logger.info = patched_logger_info
        Evaluate.__call__ = patched_evaluate_call

        try:
            yield
        finally:
            dspy.logger.info = original_logger_info
            Evaluate.__call__ = original_evaluate_call


class LangWatchTrackedMIPRO(MIPRO):
    def patch(self):
        self.metric = langwatch_dspy.track_metric(self.metric)

    def compile(self, student, **kwargs):
        with self._patch_print_and_evaluate():
            return super().compile(student, **kwargs)

    @contextmanager
    def _patch_print_and_evaluate(self):
        original_evaluate_call = Evaluate.__call__
        step = 0
        substep = 0
        scores = []

        this = self

        last_candidate_program: Optional[int] = None

        def patched_evaluate_call(self, program: dspy.Module, *args, **kwargs):
            nonlocal step, scores, substep, last_candidate_program

            if last_candidate_program != id(program):
                step += 1
                substep = 0
                scores = []
                last_candidate_program = id(program)

            step_ = str(step) if substep == 0 else f"{step}.{substep}"
            substep += 1

            score: float = original_evaluate_call(self, program, *args, **kwargs)  # type: ignore

            scores.append(score)

            if max(scores) == score:
                langwatch_dspy.log_step(
                    optimizer=DSPyOptimizer(
                        name=MIPRO.__name__,
                        parameters={
                            "num_candidates": this.num_candidates,
                            "init_temperature": this.init_temperature,
                        },
                    ),
                    index=step_,
                    score=score,
                    label="score",
                    predictors=[
                        DSPyPredictor(name=name, predictor=predictor)
                        for name, predictor in program.named_predictors()
                    ],
                )

            return score

        Evaluate.__call__ = patched_evaluate_call

        try:
            yield
        finally:
            Evaluate.__call__ = original_evaluate_call


class LangWatchTrackedMIPROv2(MIPROv2):
    def patch(self):
        self.metric = langwatch_dspy.track_metric(self.metric)

    def compile(self, student, **kwargs):
        with self._patch_print_and_evaluate():
            return super().compile(student, **kwargs)

    @contextmanager
    def _patch_print_and_evaluate(self):
        original_evaluate_call = Evaluate.__call__
        step = 0
        substep = 0
        scores = []

        this = self

        last_candidate_program: Optional[int] = None

        def patched_evaluate_call(self, program: dspy.Module, *args, **kwargs):
            nonlocal step, scores, substep, last_candidate_program

            if last_candidate_program != id(program):
                step += 1
                substep = 0
                scores = []
                last_candidate_program = id(program)

            step_ = str(step) if substep == 0 else f"{step}.{substep}"
            substep += 1

            score: float = original_evaluate_call(self, program, *args, **kwargs)  # type: ignore

            scores.append(score)

            if max(scores) == score:
                langwatch_dspy.log_step(
                    optimizer=DSPyOptimizer(
                        name=MIPROv2.__name__,
                        parameters={
                            "num_candidates": (
                                this.num_candidates
                                if hasattr(this, "num_candidates")
                                else this.n if hasattr(this, "n") else None
                            ),
                            "init_temperature": this.init_temperature,
                        },
                    ),
                    index=step_,
                    score=score,
                    label="score",
                    predictors=[
                        DSPyPredictor(name=name, predictor=predictor)
                        for name, predictor in program.named_predictors()
                    ],
                )

            return score

        Evaluate.__call__ = patched_evaluate_call

        try:
            yield
        finally:
            Evaluate.__call__ = original_evaluate_call


# === Tracer ===#


class DSPyTracer:
    def __init__(self, trace: ContextTrace):
        self.trace = trace

        if not hasattr(dspy.Module, "__original_call__"):
            dspy.Module.__original_call__ = dspy.Module.__call__  # type: ignore
            dspy.Module.__call__ = self.patched_module_call()

        if not hasattr(dspy.Predict, "__original_forward__"):
            dspy.Predict.__original_forward__ = dspy.Predict.forward  # type: ignore
            dspy.Predict.forward = self.patched_predict_forward()

        language_model_classes = dspy.OpenAI.__subclasses__() + dspy.LM.__subclasses__()
        for lm in language_model_classes:
            if not hasattr(lm, "__original_basic_request__") and hasattr(
                lm, "basic_request"
            ):
                lm.__original_basic_request__ = lm.basic_request  # type: ignore
                lm.basic_request = self.patched_legacy_language_model_request()  # type: ignore

        if not hasattr(dspy.LM, "__original_call__"):
            dspy.LM.__original_call__ = dspy.LM.__call__  # type: ignore
            dspy.LM.__call__ = self.patched_language_model_call()

        patching_retrieve = False
        if not hasattr(dspy.Retrieve, "__original_forward__"):
            dspy.Retrieve.__original_forward__ = dspy.Retrieve.forward  # type: ignore
            patching_retrieve = True

        retrieve_classes = dspy.Retrieve.__subclasses__()
        for retrieve in retrieve_classes:
            if not hasattr(retrieve, "__original_forward__"):
                retrieve.__original_forward__ = retrieve.forward  # type: ignore
                retrieve.forward = self.patched_retrieve_forward(cls=retrieve)

        if patching_retrieve:
            dspy.Retrieve.forward = self.patched_retrieve_forward(cls=dspy.Retrieve)

    def safe_get_current_span(self):
        try:
            return langwatch.get_current_span()
        except:
            return None

    def patched_module_call(self):
        self_ = self

        @langwatch.span(ignore_missing_trace_warning=True, type="module")
        def __call__(self: dspy.Module, *args, **kwargs):
            span = self_.safe_get_current_span()
            signature = (
                self.__getattribute__("signature")
                if hasattr(self, "signature")
                else None
            )

            if span and signature and hasattr(signature, "__name__"):
                span.update(name=f"{self.__class__.__name__}({signature.__name__})")
            elif span:
                span.update(name=f"{self.__class__.__name__}.forward")

            prediction = self.__class__.__original_call__(self, *args, **kwargs)  # type: ignore

            if span and isinstance(prediction, dspy.Prediction):
                span.update(output=prediction._store)  # type: ignore
            elif span:
                span.update(output=prediction)  # type: ignore

            return prediction

        return __call__

    def patched_predict_forward(self):
        self_ = self

        @langwatch.span(ignore_missing_trace_warning=True, type="module")
        def forward(self: dspy.Predict, **kwargs):
            span = self_.safe_get_current_span()
            signature = kwargs.get("signature", self.signature)

            if span and signature and hasattr(signature, "__name__"):
                span.update(name=f"{self.__class__.__name__}({signature.__name__})")
            elif span:
                span.update(name=f"{self.__class__.__name__}.forward")

            prediction = self.__class__.__original_forward__(self, **kwargs)  # type: ignore

            if span and isinstance(prediction, dspy.Prediction):
                span.update(output=prediction._store)  # type: ignore
            elif span:
                span.update(output=prediction)  # type: ignore

            return prediction

        return forward

    def patched_language_model_call(self):
        self_ = self

        @langwatch.span(ignore_missing_trace_warning=True, type="llm")
        def call(self: dspy.LM, prompt=None, messages=None, **kwargs):
            all_kwargs = self.kwargs | kwargs
            model = self.model
            params = {}
            if "temperature" in all_kwargs:
                params["temperature"] = all_kwargs["temperature"]
            if "max_tokens" in all_kwargs:
                params["max_tokens"] = all_kwargs["max_tokens"]

            span = self_.safe_get_current_span()
            if span:
                span.update(
                    name=model,
                    model=model,
                    input=(
                        messages if messages else [{"role": "user", "content": prompt}]
                    ),
                    params=params,
                )

            result = self.__class__.__original_call__(self, prompt, messages, **kwargs)  # type: ignore

            if span:
                span.update(output=result)

            history = self.history[-1] if len(self.history) > 0 else None
            if history and "usage" in history and span:
                span.update(
                    metrics={
                        "completion_tokens": history["usage"]["completion_tokens"],
                        "prompt_tokens": history["usage"]["prompt_tokens"],
                    }
                )

            return result

        return call

    def patched_legacy_language_model_request(self):
        self_ = self

        @langwatch.span(ignore_missing_trace_warning=True, type="llm")
        def basic_request(self: dspy.LM, prompt, **kwargs):
            all_kwargs = self.kwargs | kwargs
            model = all_kwargs.get("model", None)
            temperature = all_kwargs.get("temperature", None)

            span = self_.safe_get_current_span()
            if span:
                span.update(
                    name=self.__class__.__name__,
                    model=model,
                    input=prompt,
                    params=({"temperature": temperature} if temperature else None),
                )

            result = self.__class__.__original_basic_request__(self, prompt, **kwargs)  # type: ignore

            if (
                span
                and "choices" in result
                and len(result["choices"]) == 1
                and "message" in result["choices"][0]
            ):
                span.update(output=[result["choices"][0]["message"]])

            if (
                span
                and "usage" in result
                and "completion_tokens" in result["usage"]
                and "prompt_tokens" in result["usage"]
            ):
                span.update(
                    metrics={
                        "completion_tokens": result["usage"]["completion_tokens"],
                        "prompt_tokens": result["usage"]["prompt_tokens"],
                    }
                )

            return result

        return basic_request

    def patched_retrieve_forward(self, cls: Type[dspy.Retrieve]):
        self_ = self

        def forward(self: dspy.Retrieve, *args, **kwargs):
            # Prevent duplicate instrumentation
            if self.__class__ is not cls and getattr(
                cls, "forward", None
            ) is not getattr(dspy.Retrieve, "forward", None):
                return self.__class__.__original_forward__(self, *args, **kwargs)  # type: ignore

            @langwatch.span(ignore_missing_trace_warning=True, type="rag")
            def forward(self, *args, **kwargs):
                result = self.__class__.__original_forward__(self, *args, **kwargs)  # type: ignore

                span = self_.safe_get_current_span()

                passages = result
                try:
                    passages = (
                        result
                        if isinstance(result, list)
                        else result.get("passages", None)
                    )
                except:
                    warnings.warn(
                        f"LangWatch DSPy tracing: passages could not be extracted from the result for {self.__class__.__name__}, please report it at https://github.com/langwatch/langwatch/issues"
                    )
                if span and passages and type(passages) == list:
                    span.update(contexts=passages)

                if span and isinstance(result, dspy.Prediction):
                    span.update(output=result._store)  # type: ignore
                elif span:
                    span.update(output=result)

                return result

            return forward(self, *args, **kwargs)

        return forward



"""
LangWatch

This is the top level module for [LangWatch](https://github.com/langwatch/langwatch),
it holds the api_key and the endpoint the tracing is going to be send to.

For LLM and library-specific tracing functions, check out the other files on this module.
"""

import os
from warnings import warn

from langwatch.tracer import (
    trace,
    span,
    get_current_trace,
    get_current_span,
    create_span,
    capture_rag,
)
from langwatch.login import login
import langwatch.evaluations as evaluations
import langwatch.guardrails as guardrails

endpoint = os.environ.get("LANGWATCH_ENDPOINT") or "https://app.langwatch.ai"
api_key = os.environ.get("LANGWATCH_API_KEY")
debug = False

dspy_available = False
try:
    import dspy

    dspy_available = True
except AttributeError as err:
    import pydantic

    if pydantic.__version__.startswith("1."):
        warn(
            "LangWatch detected installed DSPy, however DSPy is not compatible with pydantic 1.x. Please upgrade to pydantic 2.x to use LangWatch DSPy."
        )
    else:
        raise err
except ImportError:
    pass

dspy = None
if dspy_available:
    try:
        from langwatch.dspy import langwatch_dspy

        dspy = langwatch_dspy
    except ImportError:
        warn(
            "DSPy seems to be installed but we couldn't import langwatch.dspy, please check your dspy dependency installation."
        )

langchain_available = False
try:
    import langchain

    langchain_available = True
except ImportError:
    pass

langchain = None
if langchain_available:
    try:
        import langwatch.langchain as langchain
    except ImportError:
        warn(
            "LangChain seems to be installed but we couldn't import langwatch.langchain, please check your langchain dependency installation."
        )

enabled: bool = True
sampling_rate: float = 1.0

__all__ = (
    "endpoint",
    "api_key",
    "enabled",
    "sampling_rate",
    "trace",
    "span",
    "get_current_trace",
    "get_current_span",
    "create_span",
    "capture_rag",
    "langchain",
    "dspy",
    "login",
    "evaluations",
    "guardrails",
)



import asyncio
from unittest.mock import patch
import pytest
from pytest_httpx import HTTPXMock
import requests_mock

from openai import AsyncOpenAI
import langwatch
import langwatch.guardrails
import langwatch.openai
from tests.utils import create_openai_chat_completion_mock


langwatch.api_key = "test"


class TestGuardrailsIntegration:
    @pytest.mark.asyncio
    async def test_guardrails_and_openai_integration(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        requests_mock.post(
            langwatch.endpoint + "/api/evaluations/azure-jailbreak-detection/evaluate",
            json={"status": "processed", "passed": True},
        )

        httpx_mock.add_response(
            json=create_openai_chat_completion_mock("Hello, how can I help you?"),
            url="https://api.openai.com/v1/chat/completions",
        )

        client = AsyncOpenAI(api_key="test")
        message_content = "Test message"
        with langwatch.openai.OpenAITracer(client):
            jailbreak_guardrail = await langwatch.guardrails.async_evaluate(
                "azure-jailbreak-detection", input=message_content
            )
            print("\n\njailbreak_guardrail", jailbreak_guardrail, "\n\n")
            assert jailbreak_guardrail.passed

            completion = await client.chat.completions.create(
                model="gpt-3.5-turbo",
                messages=[{"role": "user", "content": message_content}],
                stream=True,
            )

            async for part in completion:
                assert part.choices[0].delta.content == "Hello, how can I help you?"



import asyncio
from dotenv import load_dotenv

load_dotenv()

import chainlit as cl
from openai import AsyncOpenAI

client = AsyncOpenAI()

import langwatch.openai
import langwatch.guardrails


@cl.on_message
@langwatch.trace()
async def main(message: cl.Message):
    langwatch.get_current_trace().autotrack_openai_calls(client)

    msg = cl.Message(
        content="",
    )

    jailbreak_guardrail = asyncio.create_task(
        langwatch.get_current_span().async_evaluate(
            "jailbreak-detection", as_guardrail=True, input=message.content
        )
    )
    off_topic_guardrail = asyncio.create_task(
        langwatch.get_current_span().async_evaluate(
            "off-topic-evaluator", as_guardrail=True, input=message.content
        )
    )

    async def has_guardrails_failed():
        if jailbreak_guardrail.done():
            result = await jailbreak_guardrail
            if not result.passed:
                await msg.stream_token(f"I'm sorry, I can't help you with that.")
                await msg.update()
                return True

        if off_topic_guardrail.done():
            result = await off_topic_guardrail
            if not result.passed:
                await msg.stream_token(f"I'm sorry, I can't help you with that.")
                await msg.update()
                return True

        return False

    completion = await client.chat.completions.create(
        model="gpt-4",
        messages=[
            {
                "role": "system",
                "content": "You are a helpful assistant",  # that only reply in short tweet-like responses, using lots of emojis.",
            },
            {"role": "user", "content": message.content},
        ],
        stream=True,
    )

    accumulated = []
    async for part in completion:
        if await has_guardrails_failed():
            return

        if token := part.choices[0].delta.content or "":
            accumulated.append(token)

        if off_topic_guardrail.done() and jailbreak_guardrail.done():
            for token in accumulated:
                await msg.stream_token(token)
            accumulated = []

    await asyncio.gather(jailbreak_guardrail, off_topic_guardrail)
    if await has_guardrails_failed():
        return
    for token in accumulated:
        await msg.stream_token(token)

    await msg.update()



import time
import uuid
from dotenv import load_dotenv
import threading
from concurrent.futures import ThreadPoolExecutor, as_completed
import contextvars
import chainlit as cl
import langwatch

load_dotenv()

# langwatch.debug = True


@langwatch.trace()
@langwatch.span(type="llm")
def generate(question_id: str, contexts: list[str], message: str):
    langwatch.get_current_trace().update(trace_id=question_id)

    time.sleep(1)  # generating the message...
    generated_message = "Hello there! How can I help?"
    langwatch.get_current_span().update(model="custom_model")
    return generated_message


@langwatch.trace()
@langwatch.span(type="rag")
def retrieve(question_id: str, message: str):
    langwatch.get_current_trace().update(trace_id=question_id)

    time.sleep(0.5)
    contexts = ["context1", "context2"]
    langwatch.get_current_span().update(contexts=contexts)
    return contexts


@langwatch.span()
def parallel_rag(question_id: str, message: str):
    langwatch.get_current_trace().update(trace_id=question_id)

    # Create a ThreadPoolExecutor to run tasks in parallel
    with ThreadPoolExecutor(max_workers=2) as executor:
        # Submit retrieve task
        retrieve_future = executor.submit(retrieve, question_id, message)

        # Wait for retrieve to complete and then submit generate
        contexts = retrieve_future.result()
        generate_future = executor.submit(generate, question_id, contexts, message)

        # Get the final result
        return generate_future.result()


@cl.on_message
@langwatch.trace()
async def main(message: cl.Message):
    msg = cl.Message(content="")

    question_id = str(uuid.uuid4())
    # Since this is already in an async context, we can run the synchronous
    # parallel_rag in a thread pool to not block
    generated_message = await cl.make_async(parallel_rag)(question_id, message.content)

    await msg.stream_token(generated_message)
    await msg.update()



import asyncio
import sys

sys.path.insert(0, "../../")

import pydantic
import langchain
import openai

print("pydantic version:", pydantic.__version__)
print("langchain version:", langchain.__version__)
print("openai version:", openai.__version__)

if pydantic.__version__ != "1.10.15":
    raise ValueError("pydantic version must be 1.10.15")
if langchain.__version__ != "0.0.323":
    raise ValueError("langchain version must be 0.0.323")
if openai.__version__ != "0.28.1":
    raise ValueError("openai version must be 0.28.1")

from dotenv import load_dotenv
from langchain.chat_models.openai import ChatOpenAI

load_dotenv()

from langchain.prompts import ChatPromptTemplate
from langchain.schema import StrOutputParser
from langchain.schema.runnable import Runnable
from langchain.schema.runnable.config import RunnableConfig

import langwatch


model = ChatOpenAI(streaming=True)
prompt = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            "You are a helpful assistant that only reply in short tweet-like responses, using lots of emojis.",
        ),
        ("human", "{question}"),
    ]
)
runnable = prompt | model | StrOutputParser()


@langwatch.trace()
async def main(input: str):
    langwatch.get_current_trace().update(
        metadata={"customer_id": "customer_example", "labels": ["v1.0.0"]}
    )

    async for chunk in runnable.astream(
        {"question": input},
        config=RunnableConfig(
            callbacks=[
                langwatch.get_current_trace().get_langchain_callback(),
            ]
        ),
    ):
        print(chunk, end="", flush=True)

    print("\n")
    print("legacy_langchain_pydantic_bot.py:", langwatch.get_current_trace().share())


asyncio.run(main("hello"))



from dotenv import load_dotenv

load_dotenv()

import asyncio
from datetime import datetime
import time
from freezegun import freeze_time
from unittest.mock import patch
import pytest
from openai import OpenAI, AsyncOpenAI
import langwatch
import langwatch.openai
import requests_mock

from tests.utils import *  # type: ignore
from pytest_httpx import HTTPXMock
from openai.types.chat import ChatCompletionToolParam

client = OpenAI()
async_client = AsyncOpenAI()


class TestOpenAICompletionTracer:
    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_session_captures_openai_calls(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        openai_mocks = [
            create_openai_completion_mock(" there"),
            create_openai_completion_mock(" bar"),
            create_openai_completion_mock(" you!"),
            create_openai_completion_mock(" ah!"),
        ]
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_callback(
            one_mock_at_a_time(openai_mocks),
            url="https://api.openai.com/v1/completions",
        )

        with langwatch.openai.OpenAICompletionTracer(
            client,
            metadata={
                "user_id": "user-123",
                "thread_id": "thread-456",
                "customer_id": "customer-789",
            },
        ):
            response = client.completions.create(
                model="gpt-3.5-turbo-instruct", prompt="hi"
            )
            assert response.model_dump() == openai_mocks[0]
            response = client.completions.create(
                model="gpt-3.5-turbo-instruct", prompt="foo"
            )
            assert response.model_dump() == openai_mocks[1]

        time.sleep(0.01)
        trace_request = requests_mock.request_history[0].json()

        assert trace_request["metadata"]["user_id"] == "user-123"
        assert trace_request["metadata"]["thread_id"] == "thread-456"
        assert trace_request["metadata"]["customer_id"] == "customer-789"

        first_span, second_span = trace_request["spans"]

        assert first_span["trace_id"].startswith("trace_")
        assert first_span["vendor"] == "openai"
        assert first_span["model"] == "gpt-3.5-turbo-instruct"
        assert first_span["input"] == {"type": "text", "value": "hi"}
        assert first_span["outputs"] == [
            {
                "type": "text",
                "value": " there",
            }
        ]
        assert first_span["params"] == {"temperature": 1, "stream": False}
        assert first_span["metrics"] == {
            "prompt_tokens": 5,
            "completion_tokens": 16,
        }
        assert first_span["timestamps"]["started_at"] == int(
            datetime(2022, 1, 1, 0, 0, 0).timestamp() * 1000
        )
        assert first_span["timestamps"]["finished_at"] >= int(
            datetime(2022, 1, 1, 0, 0, 15).timestamp() * 1000
        )

        assert second_span["trace_id"] == first_span["trace_id"]

        with langwatch.openai.OpenAICompletionTracer(client):
            client.completions.create(
                model="gpt-3.5-turbo-instruct", prompt="we will we will rock"
            )
            client.completions.create(
                model="gpt-3.5-turbo-instruct", prompt="heey hey baby uh"
            )

        time.sleep(0.01)
        third_span = requests_mock.request_history[1].json()["spans"][0]
        assert third_span["trace_id"] != first_span["trace_id"]

        fourth_span = requests_mock.request_history[1].json()["spans"][1]
        assert fourth_span["trace_id"] == third_span["trace_id"]

    def test_trace_session_captures_exceptions(
        self, requests_mock: requests_mock.Mocker
    ):
        with patch.object(
            client.completions,
            "create",
            side_effect=Exception("An error occurred!"),
        ):
            requests_mock.post(langwatch.endpoint + "/api/collector", json={})

            with pytest.raises(Exception) as err:
                with langwatch.openai.OpenAICompletionTracer(client):
                    client.completions.create(
                        model="gpt-3.5-turbo-instruct", prompt="hi"
                    )
            assert str(err.value) == "An error occurred!"

            time.sleep(0.01)
            traced = requests_mock.request_history[0].json()["spans"][0]
            assert traced["error"]["message"] == "An error occurred!"
            assert len(traced["error"]["stacktrace"]) > 0

    @pytest.mark.asyncio
    async def test_trace_session_captures_openai_async_calls(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        openai_mocks = [
            create_openai_completion_mock(" there"),
        ]
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_callback(
            one_mock_at_a_time(openai_mocks),
            url="https://api.openai.com/v1/completions",
        )

        with langwatch.openai.OpenAICompletionTracer(async_client):
            response = await async_client.completions.create(
                model="gpt-3.5-turbo-instruct", prompt="hi"
            )
            assert response.model_dump() == openai_mocks[0]

        await asyncio.sleep(0.1)
        first_span = requests_mock.request_history[0].json()["spans"][0]
        assert first_span["outputs"] == [
            {
                "type": "text",
                "value": " there",
            }
        ]

    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_session_captures_openai_streams(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_response(
            stream=create_openai_completion_stream_mock(
                [" there", " all", " good?"], [" how", " are", " you"]
            ),
            url="https://api.openai.com/v1/completions",
        )

        with langwatch.openai.OpenAICompletionTracer(client):
            response = client.completions.create(
                model="gpt-3.5-turbo-instruct", prompt="hi", stream=True, n=2
            )
            texts = []
            for chunk in response:
                texts.append(chunk.choices[0].text)
            assert texts == [" there", " all", " good?", " how", " are", " you"]

        time.sleep(0.01)
        first_span = requests_mock.request_history[0].json()["spans"][0]
        assert first_span["outputs"] == [
            {
                "type": "text",
                "value": " there all good?",
            },
            {
                "type": "text",
                "value": " how are you",
            },
        ]
        assert first_span["vendor"] == "openai"
        assert first_span["model"] == "gpt-3.5-turbo-instruct"
        assert first_span["params"] == {"temperature": 1, "stream": True}
        assert first_span["timestamps"]["started_at"] == int(
            datetime(2022, 1, 1, 0, 0, 0).timestamp() * 1000
        )
        assert first_span["timestamps"]["first_token_at"] >= int(
            datetime(2022, 1, 1, 0, 0, 15).timestamp() * 1000
        )
        assert first_span["timestamps"]["finished_at"] >= int(
            datetime(2022, 1, 1, 0, 0, 30).timestamp() * 1000
        )

    @pytest.mark.asyncio
    async def test_trace_session_captures_openai_async_streams(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_response(
            stream=create_openai_completion_stream_mock(
                [" there", " all", " good?"], [" how", " are", " you"]
            ),
            url="https://api.openai.com/v1/completions",
        )

        with langwatch.openai.OpenAICompletionTracer(async_client):
            response = await async_client.completions.create(
                model="gpt-3.5-turbo-instruct", prompt="hi", stream=True, n=2
            )
            texts = []
            async for chunk in response:
                texts.append(chunk.choices[0].text)
            assert texts == [" there", " all", " good?", " how", " are", " you"]

        time.sleep(0.01)
        first_span = requests_mock.request_history[0].json()["spans"][0]
        assert first_span["outputs"] == [
            {
                "type": "text",
                "value": " there all good?",
            },
            {
                "type": "text",
                "value": " how are you",
            },
        ]
        assert first_span["vendor"] == "openai"
        assert first_span["model"] == "gpt-3.5-turbo-instruct"
        assert first_span["params"] == {"temperature": 1, "stream": True}

    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_nested_spans(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_callback(
            one_mock_at_a_time(
                [
                    create_openai_completion_mock(" there"),
                    create_openai_completion_mock(" are"),
                    create_openai_completion_mock("?"),
                ]
            ),
            url="https://api.openai.com/v1/completions",
        )

        with langwatch.openai.OpenAICompletionTracer(client):
            client.completions.create(model="gpt-3.5-turbo-instruct", prompt="hi")
            with langwatch.create_span("subtask"):
                client.completions.create(model="gpt-3.5-turbo-instruct", prompt="how")
                client.completions.create(model="gpt-3.5-turbo-instruct", prompt="you")

        time.sleep(0.01)

        (
            first_span,
            second_span,
            third_span,
            fourth_span,
        ) = requests_mock.request_history[0].json()["spans"]

        assert first_span["type"] == "llm"
        assert first_span["trace_id"].startswith("trace_")
        assert first_span["span_id"].startswith("span_")
        assert first_span["parent_id"] == None

        assert fourth_span["type"] == "span"
        assert fourth_span["name"] == "subtask"
        assert fourth_span["trace_id"] == first_span["trace_id"]
        assert fourth_span["span_id"].startswith("span_")
        assert fourth_span["parent_id"] == None

        assert second_span["type"] == "llm"
        assert second_span["trace_id"] == first_span["trace_id"]
        assert second_span["parent_id"] == fourth_span["span_id"]

        assert third_span["type"] == "llm"
        assert third_span["trace_id"] == first_span["trace_id"]
        assert third_span["parent_id"] == fourth_span["span_id"]

    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_nested_spans_using_annotations(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_callback(
            one_mock_at_a_time(
                [
                    create_openai_completion_mock(" there"),
                    create_openai_completion_mock(" are"),
                    create_openai_completion_mock("?"),
                ]
            ),
            url="https://api.openai.com/v1/completions",
        )

        @langwatch.span()
        def subtask(foo=None):
            client.completions.create(model="gpt-3.5-turbo-instruct", prompt="how")
            client.completions.create(model="gpt-3.5-turbo-instruct", prompt="you")

            return foo

        with langwatch.openai.OpenAICompletionTracer(client):
            client.completions.create(model="gpt-3.5-turbo-instruct", prompt="hi")
            subtask(foo="bar")

        time.sleep(0.01)

        (
            first_span,
            second_span,
            third_span,
            fourth_span,
        ) = requests_mock.request_history[0].json()["spans"]

        assert first_span["type"] == "llm"
        assert first_span["trace_id"].startswith("trace_")
        assert first_span["span_id"].startswith("span_")
        assert first_span["parent_id"] == None

        assert fourth_span["type"] == "span"
        assert fourth_span["name"] == "subtask"
        assert fourth_span["input"] == {"type": "json", "value": {"foo": "bar"}}
        assert fourth_span["outputs"] == [{"type": "text", "value": "bar"}]
        assert fourth_span["trace_id"] == first_span["trace_id"]
        assert fourth_span["span_id"].startswith("span_")
        assert fourth_span["parent_id"] == None

        assert second_span["type"] == "llm"
        assert second_span["trace_id"] == first_span["trace_id"]
        assert second_span["parent_id"] == fourth_span["span_id"]

        assert third_span["type"] == "llm"
        assert third_span["trace_id"] == first_span["trace_id"]
        assert third_span["parent_id"] == fourth_span["span_id"]


class TestOpenAIChatCompletionTracer:
    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_session_captures_openai_calls(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        openai_mocks = [
            create_openai_chat_completion_mock("hi there!"),
            create_openai_chat_completion_mock("bar baz"),
        ]
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_callback(
            one_mock_at_a_time(openai_mocks),
            url="https://api.openai.com/v1/chat/completions",
        )

        with langwatch.openai.OpenAIChatCompletionTracer(
            client,
            metadata={
                "user_id": "user-123",
                "thread_id": "thread-456",
                "customer_id": "customer-789",
                "labels": ["1.0.0"],
            },
        ):
            response = client.chat.completions.create(
                model="gpt-3.5-turbo",
                messages=[{"role": "user", "content": "Hello!"}],
            )
            assert response.model_dump() == openai_mocks[0]
            response = client.chat.completions.create(
                model="gpt-3.5-turbo",
                messages=[{"role": "user", "content": "foo"}],
            )
            assert response.model_dump() == openai_mocks[1]

        time.sleep(0.01)
        trace_request = requests_mock.request_history[0].json()

        assert trace_request["trace_id"].startswith("trace_")
        assert trace_request["metadata"]["user_id"] == "user-123"
        assert trace_request["metadata"]["thread_id"] == "thread-456"
        assert trace_request["metadata"]["customer_id"] == "customer-789"
        assert trace_request["metadata"]["labels"] == ["1.0.0"]

        first_span = trace_request["spans"][0]
        assert first_span["trace_id"].startswith("trace_")
        assert first_span["vendor"] == "openai"
        assert first_span["model"] == "gpt-3.5-turbo"
        assert first_span["input"] == {
            "type": "chat_messages",
            "value": [
                {
                    "role": "user",
                    "content": "Hello!",
                }
            ],
        }
        assert first_span["outputs"] == [
            {
                "type": "chat_messages",
                "value": [
                    {
                        "role": "assistant",
                        "content": "hi there!",
                        "function_call": None,
                        "tool_calls": None,
                    }
                ],
            }
        ]
        assert first_span["params"] == {"temperature": 1, "stream": False}
        assert first_span["metrics"] == {
            "prompt_tokens": 5,
            "completion_tokens": 16,
        }
        assert first_span["timestamps"]["started_at"] == int(
            datetime(2022, 1, 1, 0, 0, 0).timestamp() * 1000
        )
        assert first_span["timestamps"]["finished_at"] >= int(
            datetime(2022, 1, 1, 0, 0, 15).timestamp() * 1000
        )

        second_span = requests_mock.request_history[0].json()["spans"][1]
        assert second_span["trace_id"] == first_span["trace_id"]

    def test_trace_session_captures_exceptions(
        self, requests_mock: requests_mock.Mocker
    ):
        with patch.object(
            client.chat.completions,
            "create",
            side_effect=Exception("An error occurred!"),
        ):
            requests_mock.post(langwatch.endpoint + "/api/collector", json={})

            with pytest.raises(Exception) as err:
                with langwatch.openai.OpenAIChatCompletionTracer(client):
                    client.chat.completions.create(model="gpt-3.5-turbo", messages=[])
            assert str(err.value) == "An error occurred!"

            time.sleep(0.01)
            traced = requests_mock.request_history[0].json()["spans"][0]
            assert traced["error"]["message"] == "An error occurred!"
            assert len(traced["error"]["stacktrace"]) > 0

    @pytest.mark.asyncio
    async def test_trace_session_captures_openai_async_calls(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        openai_mocks = [
            create_openai_chat_completion_mock("hi there!"),
        ]
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_callback(
            one_mock_at_a_time(openai_mocks),
            url="https://api.openai.com/v1/chat/completions",
        )

        with langwatch.openai.OpenAIChatCompletionTracer(async_client):
            response = await async_client.chat.completions.create(
                model="gpt-3.5-turbo",
                messages=[{"role": "user", "content": "Hello!"}],
            )
            assert response.model_dump() == openai_mocks[0]

        await asyncio.sleep(0.1)
        first_span = requests_mock.request_history[0].json()["spans"][0]
        assert first_span["outputs"] == [
            {
                "type": "chat_messages",
                "value": [
                    {
                        "role": "assistant",
                        "content": "hi there!",
                        "function_call": None,
                        "tool_calls": None,
                    }
                ],
            }
        ]

    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_session_captures_openai_function_calls(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        openai_mocks = [
            {
                "id": "chatcmpl-86zIvz53Wa4qTc1ksUt3coF5yTvm7",
                "object": "chat.completion",
                "created": 1696676313,
                "model": "gpt-3.5-turbo-0613",
                "choices": [
                    {
                        "index": 0,
                        "message": {
                            "role": "assistant",
                            "function_call": {
                                "arguments": '{\n  "input": "2+2"\n}',
                                "name": "Calculator",
                            },
                            "content": None,
                        },
                        "finish_reason": "function_call",
                    }
                ],
                "usage": {
                    "prompt_tokens": 5,
                    "completion_tokens": 16,
                    "total_tokens": 21,
                },
            }
        ]
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_callback(
            one_mock_at_a_time(openai_mocks),
            url="https://api.openai.com/v1/chat/completions",
        )

        with langwatch.openai.OpenAIChatCompletionTracer(client):
            client.chat.completions.create(
                model="gpt-3.5-turbo",
                messages=[{"role": "user", "content": "how much is 2 + 2?"}],
                functions=[
                    {
                        "name": "Calculator",
                        "description": "useful for when you need to answer questions about math",
                        "parameters": {
                            "properties": {
                                "expression": {
                                    "title": "expression",
                                    "type": "string",
                                }
                            },
                            "required": ["expression"],
                            "type": "object",
                        },
                    }
                ],
            )

        time.sleep(0.01)
        first_span = requests_mock.request_history[0].json()["spans"][0]
        assert first_span["trace_id"].startswith("trace_")
        assert first_span["vendor"] == "openai"
        assert first_span["model"] == "gpt-3.5-turbo"
        assert first_span["input"] == {
            "type": "chat_messages",
            "value": [{"role": "user", "content": "how much is 2 + 2?"}],
        }
        assert first_span["outputs"] == [
            {
                "type": "chat_messages",
                "value": [
                    {
                        "role": "assistant",
                        "content": None,
                        "function_call": {
                            "name": "Calculator",
                            "arguments": '{\n  "input": "2+2"\n}',
                        },
                        "tool_calls": None,
                    }
                ],
            }
        ]
        assert first_span["params"] == {
            "temperature": 1.0,
            "stream": False,
            "functions": [
                {
                    "name": "Calculator",
                    "description": "useful for when you need to answer questions about math",
                    "parameters": {
                        "properties": {
                            "expression": {"title": "expression", "type": "string"}
                        },
                        "required": ["expression"],
                        "type": "object",
                    },
                }
            ],
        }

    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_session_captures_openai_streams(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_response(
            stream=create_openai_chat_completion_stream_mock(
                ["Hi", " there", " all", " good?"], ["Hi", " how", " are", " you"]
            ),
            url="https://api.openai.com/v1/chat/completions",
        )

        with langwatch.openai.OpenAIChatCompletionTracer(client):
            response = client.chat.completions.create(
                model="gpt-3.5-turbo",
                messages=[{"role": "user", "content": "Hello!"}],
                stream=True,
                n=2,
            )
            texts = []
            for chunk in response:
                texts.append(chunk.choices[0].delta.content)
            assert texts == [
                "",
                "",
                "Hi",
                " there",
                " all",
                " good?",
                "Hi",
                " how",
                " are",
                " you",
                None,
                None,
            ]

        time.sleep(0.01)
        first_span = requests_mock.request_history[0].json()["spans"][0]
        assert first_span["outputs"] == [
            {
                "type": "chat_messages",
                "value": [{"role": "assistant", "content": "Hi there all good?"}],
            },
            {
                "type": "chat_messages",
                "value": [{"role": "assistant", "content": "Hi how are you"}],
            },
        ]
        assert first_span["vendor"] == "openai"
        assert first_span["model"] == "gpt-3.5-turbo"
        assert first_span["params"] == {"temperature": 1, "stream": True}
        assert first_span["timestamps"]["started_at"] == int(
            datetime(2022, 1, 1, 0, 0, 0).timestamp() * 1000
        )
        assert first_span["timestamps"]["first_token_at"] >= int(
            datetime(2022, 1, 1, 0, 0, 15).timestamp() * 1000
        )
        assert first_span["timestamps"]["finished_at"] >= int(
            datetime(2022, 1, 1, 0, 0, 30).timestamp() * 1000
        )

    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_session_captures_openai_streams_with_functions(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_response(
            stream=create_openai_chat_completion_function_stream_mock(
                {"name": "Calculator", "arguments": '{\n  "input": "2+2"\n}'}
            ),
            url="https://api.openai.com/v1/chat/completions",
        )

        with langwatch.openai.OpenAIChatCompletionTracer(client):
            response = client.chat.completions.create(
                model="gpt-3.5-turbo",
                messages=[{"role": "user", "content": "how much is 2 + 2?"}],
                stream=True,
                functions=[
                    {
                        "name": "Calculator",
                        "description": "useful for when you need to answer questions about math",
                        "parameters": {
                            "properties": {
                                "expression": {
                                    "title": "expression",
                                    "type": "string",
                                }
                            },
                            "required": ["expression"],
                            "type": "object",
                        },
                    }
                ],
            )
            for _ in response:
                pass  # just to consume the stream

        time.sleep(0.01)
        first_span = requests_mock.request_history[0].json()["spans"][0]
        assert first_span["outputs"] == [
            {
                "type": "chat_messages",
                "value": [
                    {
                        "role": "assistant",
                        "function_call": {
                            "arguments": '{\n  "input": "2+2"\n}',
                            "name": "Calculator",
                        },
                        "content": None,
                    }
                ],
            }
        ]

    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_session_captures_openai_tool_calls(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        openai_mocks = [
            {
                "id": "chatcmpl-8Sf6QZXVnDyjAIQcmjcRiYeMKY49D",
                "object": "chat.completion",
                "created": 1701841874,
                "model": "gpt-3.5-turbo-0613",
                "choices": [
                    {
                        "index": 0,
                        "message": {
                            "role": "assistant",
                            "content": None,
                            "tool_calls": [
                                {
                                    "id": "call_hF48nuhUAwKVFQ3JZKdcSDZU",
                                    "type": "function",
                                    "function": {
                                        "name": "get_current_weather",
                                        "arguments": '{\n  "location": "Boston"\n}',
                                    },
                                }
                            ],
                        },
                        "finish_reason": "tool_calls",
                    }
                ],
                "usage": {
                    "prompt_tokens": 82,
                    "completion_tokens": 16,
                    "total_tokens": 98,
                },
                "system_fingerprint": None,
            }
        ]
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_callback(
            one_mock_at_a_time(openai_mocks),
            url="https://api.openai.com/v1/chat/completions",
        )

        tools: List[ChatCompletionToolParam] = [
            {
                "type": "function",
                "function": {
                    "name": "get_current_weather",
                    "description": "Get the current weather in a given location",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "location": {
                                "type": "string",
                                "description": "The city and state, e.g. San Francisco, CA",
                            },
                            "unit": {
                                "type": "string",
                                "enum": ["celsius", "fahrenheit"],
                            },
                        },
                        "required": ["location"],
                    },
                },
            }
        ]
        with langwatch.openai.OpenAIChatCompletionTracer(client):
            client.chat.completions.create(
                model="gpt-3.5-turbo",
                messages=[
                    {"role": "user", "content": "What is the weather like in Boston?"}
                ],
                tools=tools,
                tool_choice="auto",
            )

        time.sleep(0.01)
        first_span = requests_mock.request_history[0].json()["spans"][0]
        assert first_span["params"]["tools"] == tools
        assert first_span["params"]["tool_choice"] == "auto"
        assert first_span["outputs"] == [
            {
                "type": "chat_messages",
                "value": [
                    {
                        "role": "assistant",
                        "content": None,
                        "function_call": None,
                        "tool_calls": [
                            {
                                "id": "call_hF48nuhUAwKVFQ3JZKdcSDZU",
                                "type": "function",
                                "function": {
                                    "name": "get_current_weather",
                                    "arguments": '{\n  "location": "Boston"\n}',
                                },
                            }
                        ],
                    }
                ],
            }
        ]

    @freeze_time("2022-01-01", auto_tick_seconds=15)
    def test_trace_session_captures_openai_streams_with_tools(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_response(
            stream=IteratorStream(
                [
                    str.encode(line + "\n\n")
                    for line in """data: {"id":"chatcmpl-8Sf3okV8aZ95sKdzYJA9bVIU8aVts","object":"chat.completion.chunk","created":1701841712,"model":"gpt-3.5-turbo-0613","system_fingerprint":null,"choices":[{"index":0,"delta":{"role":"assistant","content":null,"tool_calls":[{"index":0,"id":"call_oTTUQg7fmAiF6MlGyQ2yLrh1","type":"function","function":{"name":"get_current_weather","arguments":""}}]},"finish_reason":null}]}

data: {"id":"chatcmpl-8Sf3okV8aZ95sKdzYJA9bVIU8aVts","object":"chat.completion.chunk","created":1701841712,"model":"gpt-3.5-turbo-0613","system_fingerprint":null,"choices":[{"index":0,"delta":{"tool_calls":[{"index":0,"function":{"arguments":"{\\n"}}]},"finish_reason":null}]}

data: {"id":"chatcmpl-8Sf3okV8aZ95sKdzYJA9bVIU8aVts","object":"chat.completion.chunk","created":1701841712,"model":"gpt-3.5-turbo-0613","system_fingerprint":null,"choices":[{"index":0,"delta":{"tool_calls":[{"index":0,"function":{"arguments":"\\""}}]},"finish_reason":null}]}

data: {"id":"chatcmpl-8Sf3okV8aZ95sKdzYJA9bVIU8aVts","object":"chat.completion.chunk","created":1701841712,"model":"gpt-3.5-turbo-0613","system_fingerprint":null,"choices":[{"index":0,"delta":{"tool_calls":[{"index":0,"function":{"arguments":"location"}}]},"finish_reason":null}]}

data: {"id":"chatcmpl-8Sf3okV8aZ95sKdzYJA9bVIU8aVts","object":"chat.completion.chunk","created":1701841712,"model":"gpt-3.5-turbo-0613","system_fingerprint":null,"choices":[{"index":0,"delta":{"tool_calls":[{"index":0,"function":{"arguments":"\\":"}}]},"finish_reason":null}]}

data: {"id":"chatcmpl-8Sf3okV8aZ95sKdzYJA9bVIU8aVts","object":"chat.completion.chunk","created":1701841712,"model":"gpt-3.5-turbo-0613","system_fingerprint":null,"choices":[{"index":0,"delta":{"tool_calls":[{"index":0,"function":{"arguments":" \\""}}]},"finish_reason":null}]}

data: {"id":"chatcmpl-8Sf3okV8aZ95sKdzYJA9bVIU8aVts","object":"chat.completion.chunk","created":1701841712,"model":"gpt-3.5-turbo-0613","system_fingerprint":null,"choices":[{"index":0,"delta":{"tool_calls":[{"index":0,"function":{"arguments":"Boston"}}]},"finish_reason":null}]}

data: {"id":"chatcmpl-8Sf3okV8aZ95sKdzYJA9bVIU8aVts","object":"chat.completion.chunk","created":1701841712,"model":"gpt-3.5-turbo-0613","system_fingerprint":null,"choices":[{"index":0,"delta":{"tool_calls":[{"index":0,"function":{"arguments":","}}]},"finish_reason":null}]}

data: {"id":"chatcmpl-8Sf3okV8aZ95sKdzYJA9bVIU8aVts","object":"chat.completion.chunk","created":1701841712,"model":"gpt-3.5-turbo-0613","system_fingerprint":null,"choices":[{"index":0,"delta":{"tool_calls":[{"index":0,"function":{"arguments":" MA"}}]},"finish_reason":null}]}

data: {"id":"chatcmpl-8Sf3okV8aZ95sKdzYJA9bVIU8aVts","object":"chat.completion.chunk","created":1701841712,"model":"gpt-3.5-turbo-0613","system_fingerprint":null,"choices":[{"index":0,"delta":{"tool_calls":[{"index":0,"function":{"arguments":"\\"\\n"}}]},"finish_reason":null}]}

data: {"id":"chatcmpl-8Sf3okV8aZ95sKdzYJA9bVIU8aVts","object":"chat.completion.chunk","created":1701841712,"model":"gpt-3.5-turbo-0613","system_fingerprint":null,"choices":[{"index":0,"delta":{"tool_calls":[{"index":0,"function":{"arguments":"}"}}]},"finish_reason":null}]}

data: {"id":"chatcmpl-8Sf3okV8aZ95sKdzYJA9bVIU8aVts","object":"chat.completion.chunk","created":1701841712,"model":"gpt-3.5-turbo-0613","system_fingerprint":null,"choices":[{"index":0,"delta":{},"finish_reason":"tool_calls"}]}

data: [DONE]""".split(
                        "\n\n"
                    )
                ]
            ),
            url="https://api.openai.com/v1/chat/completions",
        )

        with langwatch.openai.OpenAIChatCompletionTracer(client):
            response = client.chat.completions.create(
                model="gpt-3.5-turbo",
                messages=[
                    {"role": "user", "content": "What is the weather like in Boston?"}
                ],
                stream=True,
                tools=[
                    {
                        "type": "function",
                        "function": {
                            "name": "get_current_weather",
                            "description": "Get the current weather in a given location",
                            "parameters": {
                                "type": "object",
                                "properties": {
                                    "location": {
                                        "type": "string",
                                        "description": "The city and state, e.g. San Francisco, CA",
                                    },
                                    "unit": {
                                        "type": "string",
                                        "enum": ["celsius", "fahrenheit"],
                                    },
                                },
                                "required": ["location"],
                            },
                        },
                    }
                ],
                tool_choice="auto",
            )
            for _ in response:
                pass  # just to consume the stream

        time.sleep(0.01)
        first_span = requests_mock.request_history[0].json()["spans"][0]
        assert first_span["outputs"] == [
            {
                "type": "chat_messages",
                "value": [
                    {
                        "role": "assistant",
                        "content": None,
                        "tool_calls": [
                            {
                                "id": "call_oTTUQg7fmAiF6MlGyQ2yLrh1",
                                "type": "function",
                                "function": {
                                    "name": "get_current_weather",
                                    "arguments": '{\n"location": "Boston, MA"\n}',
                                },
                            }
                        ],
                    }
                ],
            }
        ]

    @pytest.mark.asyncio
    async def test_trace_session_captures_openai_async_streams(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_response(
            stream=create_openai_chat_completion_stream_mock(
                ["Hi", " there", " all", " good?"], ["Hi", " how", " are", " you"]
            ),
            url="https://api.openai.com/v1/chat/completions",
        )

        with langwatch.openai.OpenAIChatCompletionTracer(async_client):
            response = await async_client.chat.completions.create(
                model="gpt-3.5-turbo",
                messages=[{"role": "user", "content": "Hello!"}],
                stream=True,
                n=2,
            )
            texts = []
            async for chunk in response:
                texts.append(chunk.choices[0].delta.content)
            assert texts == [
                "",
                "",
                "Hi",
                " there",
                " all",
                " good?",
                "Hi",
                " how",
                " are",
                " you",
                None,
                None,
            ]

        time.sleep(0.01)
        first_span = requests_mock.request_history[0].json()["spans"][0]
        assert first_span["outputs"] == [
            {
                "type": "chat_messages",
                "value": [{"role": "assistant", "content": "Hi there all good?"}],
            },
            {
                "type": "chat_messages",
                "value": [{"role": "assistant", "content": "Hi how are you"}],
            },
        ]
        assert first_span["model"] == "gpt-3.5-turbo"
        assert first_span["params"] == {"temperature": 1, "stream": True}

    def test_trace_rag(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        openai_mocks = [
            create_openai_chat_completion_mock("The capital of France is Paris."),
        ]
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_callback(
            one_mock_at_a_time(openai_mocks),
            url="https://api.openai.com/v1/chat/completions",
        )

        with langwatch.openai.OpenAIChatCompletionTracer(client):
            with langwatch.capture_rag(
                input="What is the capital of France?",
                contexts=[
                    {
                        "document_id": "doc-1",
                        "chunk_id": "0",
                        "content": "France is a country in Europe.",
                    },
                    {
                        "document_id": "doc-2",
                        "chunk_id": "0",
                        "content": "Paris is the capital of France.",
                    },
                ],
            ):
                response = client.chat.completions.create(
                    model="gpt-3.5-turbo",
                    messages=[
                        {"role": "user", "content": "What is the capital of France?"}
                    ],
                )

        time.sleep(0.01)

        llm_span, rag_span = requests_mock.request_history[0].json()["spans"]

        assert rag_span["type"] == "rag"
        assert rag_span["input"]["value"] == "What is the capital of France?"
        assert rag_span["contexts"] == [
            {
                "document_id": "doc-1",
                "chunk_id": "0",
                "content": "France is a country in Europe.",
            },
            {
                "document_id": "doc-2",
                "chunk_id": "0",
                "content": "Paris is the capital of France.",
            },
        ]

        assert llm_span["type"] == "llm"
        assert (
            llm_span["outputs"][0]["value"][0]["content"]
            == response.choices[0].message.content
        )


class TestOpenAITracer:
    def test_traces_both_completion_and_chat_completion(
        self, httpx_mock: HTTPXMock, requests_mock: requests_mock.Mocker
    ):
        requests_mock.post(langwatch.endpoint + "/api/collector", json={})
        httpx_mock.add_callback(
            one_mock_at_a_time([create_openai_completion_mock("foo")]),
            url="https://api.openai.com/v1/completions",
        )
        httpx_mock.add_callback(
            one_mock_at_a_time([create_openai_chat_completion_mock("bar")]),
            url="https://api.openai.com/v1/chat/completions",
        )

        with langwatch.openai.OpenAITracer(
            client,
            metadata={
                "user_id": "user-123",
                "thread_id": "thread-456",
                "customer_id": "customer-789",
            },
        ):
            client.completions.create(
                model="gpt-3.5-turbo-instruct", prompt="Hello Completion!"
            )
            client.chat.completions.create(
                model="gpt-3.5-turbo",
                messages=[{"role": "user", "content": "Hello ChatCompletion!"}],
            )

        time.sleep(0.01)
        first_span = requests_mock.request_history[0].json()["spans"][0]
        second_span = requests_mock.request_history[1].json()["spans"][0]
        assert first_span["trace_id"] == second_span["trace_id"]



import os
from dotenv import load_dotenv

load_dotenv()

import chainlit as cl

import dspy

import os
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk import trace as trace_sdk
from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SimpleSpanProcessor

from openinference.instrumentation.dspy import DSPyInstrumentor

# Set up OpenTelemetry trace provider with LangWatch as the endpoint
tracer_provider = trace_sdk.TracerProvider()
tracer_provider.add_span_processor(
    SimpleSpanProcessor(
        OTLPSpanExporter(
            endpoint=f"{os.environ.get('LANGWATCH_ENDPOINT', 'https://app.langwatch.ai')}/api/otel/v1/traces",
            headers={"Authorization": "Bearer " + os.environ["LANGWATCH_API_KEY"]},
        )
    )
)
# Optionally, you can also print the spans to the console.
tracer_provider.add_span_processor(SimpleSpanProcessor(ConsoleSpanExporter()))

DSPyInstrumentor().instrument(tracer_provider=tracer_provider)


llm = dspy.OpenAI(
    model="gpt-4o-mini",
    max_tokens=2048,
    temperature=0,
    api_key=os.environ["OPENAI_API_KEY"],
)

colbertv2_wiki17_abstracts = dspy.ColBERTv2(
    url="http://20.102.90.50:2017/wiki17_abstracts"
)

dspy.settings.configure(lm=llm, rm=colbertv2_wiki17_abstracts)


class GenerateAnswer(dspy.Signature):
    """Answer questions with short factoid answers."""

    context = dspy.InputField(desc="may contain relevant facts")
    question = dspy.InputField()
    answer = dspy.OutputField(desc="often between 1 and 5 words")


class RAG(dspy.Module):
    def __init__(self, num_passages=3):
        super().__init__()

        self.retrieve = dspy.Retrieve(k=num_passages)
        self.generate_answer = dspy.ChainOfThought(GenerateAnswer)

    def forward(self, question):
        context = self.retrieve(question).passages  # type: ignore
        prediction = self.generate_answer(question=question, context=context)
        return dspy.Prediction(answer=prediction.answer)


@cl.on_message
async def main(message: cl.Message):
    msg = cl.Message(
        content="",
    )

    program = RAG()
    program.load(
        f"{os.path.dirname(os.path.abspath(__file__))}/../data/rag_dspy_bot.json",
        use_legacy_loading=True,
    )
    program = program.reset_copy()
    prediction = program(question=message.content)

    await msg.stream_token(prediction.answer)
    await msg.update()



# This example uses the OpenTelemetry instrumentation for OpenAI from OpenLLMetry: https://pypi.org/project/opentelemetry-instrumentation-openai/

from dotenv import load_dotenv

load_dotenv()

import chainlit as cl

import os
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk import trace as trace_sdk
from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SimpleSpanProcessor

from opentelemetry.instrumentation.openai import OpenAIInstrumentor
from openai import OpenAI


client = OpenAI()


# Set up OpenTelemetry trace provider with LangWatch as the endpoint
tracer_provider = trace_sdk.TracerProvider()
tracer_provider.add_span_processor(
    SimpleSpanProcessor(
        OTLPSpanExporter(
            endpoint=f"{os.environ.get('LANGWATCH_ENDPOINT', 'https://app.langwatch.ai')}/api/otel/v1/traces",
            headers={"Authorization": "Bearer " + os.environ["LANGWATCH_API_KEY"]},
        )
    )
)
# Optionally, you can also print the spans to the console.
tracer_provider.add_span_processor(SimpleSpanProcessor(ConsoleSpanExporter()))

OpenAIInstrumentor().instrument(tracer_provider=tracer_provider)


@cl.on_message
async def main(message: cl.Message):
    msg = cl.Message(
        content="",
    )

    completion = client.chat.completions.create(
        model="gpt-4o-mini",
        messages=[
            {
                "role": "system",
                "content": "You are a helpful assistant that only reply in short tweet-like responses, using lots of emojis.",
            },
            {"role": "user", "content": message.content},
        ],
        stream=True,
    )

    for part in completion:
        if token := part.choices[0].delta.content or "":
            await msg.stream_token(token)

    await msg.update()



import os
from typing import Optional
from dotenv import load_dotenv

from langwatch.types import RAGChunk

load_dotenv()

import chainlit as cl
from openai import AzureOpenAI

import langwatch

client = AzureOpenAI(
    api_key=os.getenv("AZURE_OPENAI_API_KEY"),
    api_version="2024-02-01",
    azure_endpoint=os.getenv("AZURE_OPENAI_ENDPOINT"),  # type: ignore
)
langwatch.api_key = os.getenv("LANGWATCH_API_KEY")


@langwatch.span(type="rag")
def retrieve(query: Optional[str] = None):
    search_results = [
        {
            "id": "result_1",
            "content": "This is the first result",
        },
        {
            "id": "result_2",
            "content": "This is the second result",
        },
    ]

    langwatch.get_current_span().update(
        contexts=[
            RAGChunk(
                document_id=docs["id"],
                content=docs["content"],
            )
            for docs in search_results
        ],
    )

    return search_results


@cl.on_message
@langwatch.trace()
async def main(message: cl.Message):
    langwatch.get_current_trace().autotrack_openai_calls(client)

    msg = cl.Message(
        content="",
    )

    langwatch.get_current_trace().update(
        trace_id=message.id,
        metadata={"user_id": message.author},
    )

    completion = client.chat.completions.create(
        model="gpt-35-turbo-0613",
        messages=[
            {
                "role": "system",
                "content": "come up with a query for searching the database based on user question, 3 words max",
            },
            {"role": "user", "content": message.content},
        ],
    )

    query = completion.choices[0].message.content
    search_results = retrieve(query=query)
    results = "\n".join([f"{docs['id']}: {docs['content']}" for docs in search_results])

    completion = client.chat.completions.create(
        model="gpt-35-turbo-0613",
        messages=[
            {
                "role": "system",
                "content": f"""
                    You are a helpful assistant that only reply in short tweet-like responses, using lots of emojis.

                    We just made a search in the database for {query} and found {len(search_results)} results. Here they are, use that to help answering user:

                    {results}
                """,
            },
            {"role": "user", "content": message.content},
        ],
        stream=True,
    )

    for part in completion:
        if len(part.choices) == 0:
            continue

        if token := part.choices[0].delta.content or "":
            await msg.stream_token(token)

    await msg.update()



from dotenv import load_dotenv

load_dotenv()

from typing import Optional
import streamlit as st
from openai import OpenAI
import langwatch
from langwatch.tracer import ContextSpan

client = OpenAI()

assistant = client.beta.assistants.create(
    name="Math Tutor",
    instructions="You are a personal math tutor. Write and run code to answer math questions.",
    tools=[{"type": "code_interpreter"}],
    model="gpt-4o-mini",
)

thread = client.beta.threads.create()


@langwatch.span(capture_output=False)
def pipeline(message: str):
    langwatch.get_current_trace().autotrack_openai_calls(client)
    langwatch.get_current_trace().update(metadata={"thread_id": thread.id})

    client.beta.threads.messages.create(
        thread_id=thread.id,
        role="user",
        content=message,
    )

    llm_span: Optional[ContextSpan] = None
    step_span: Optional[ContextSpan] = None

    with client.beta.threads.runs.stream(
        thread_id=thread.id,
        assistant_id=assistant.id,
        instructions="Please address the user as Jane Doe. The user has a premium account.",
    ) as stream:
        for event in stream:
            if event.event == "thread.run.created":
                llm_span = langwatch.get_current_span().span(
                    type="llm", input=event.data.instructions, model=event.data.model
                )

            elif event.event == "thread.run.completed":
                if llm_span:
                    usage = event.data.usage
                    completion_tokens = usage.completion_tokens if usage else None
                    prompt_tokens = usage.prompt_tokens if usage else None
                    llm_span.end(
                        metrics={
                            "completion_tokens": completion_tokens,
                            "prompt_tokens": prompt_tokens,
                        }
                    )
                    llm_span = None

            elif event.event == "thread.run.step.created":
                if llm_span:
                    step_span = llm_span.span(
                        type="tool" if event.data.type == "tool_calls" else "span",
                        name=event.data.step_details.type,
                    )

            elif event.event == "thread.run.step.delta":
                if step_span:
                    if (
                        event.data.delta.step_details
                        and event.data.delta.step_details.type == "tool_calls"
                        and event.data.delta.step_details.tool_calls
                        and len(event.data.delta.step_details.tool_calls) > 0
                    ):
                        tool_call = event.data.delta.step_details.tool_calls[0]
                        if (
                            tool_call.type == "code_interpreter"
                            and tool_call.code_interpreter
                        ):
                            step_span.update(
                                name=tool_call.type,
                                input=(step_span.input or "") + (tool_call.code_interpreter.input or ""),  # type: ignore
                                output=(step_span.output or []) + (tool_call.code_interpreter.outputs or []),  # type: ignore
                            )
                        # Add other tool calls here

            elif event.event == "thread.run.step.completed":
                if step_span:
                    step_span.end()
                    step_span = None

            elif event.data.object == "thread.message.delta" and event.data.delta.content:  # type: ignore
                text = event.data.delta.content[0].text.value or ""  # type: ignore

                if llm_span:
                    llm_span.update(output=(llm_span.output or "") + text)  # type: ignore

                if step_span:
                    step_span.update(output=(step_span.output or "") + text)  # type: ignore

                yield text

            else:
                pass


@langwatch.trace()
def process_message():
    message = st.session_state.prompt
    with st.chat_message("user"):
        st.markdown(message)

    with st.chat_message("assistant"):
        message_placeholder = st.empty()

        message_placeholder.write_stream(pipeline(message))


prompt = st.chat_input(
    "Ask your question here: ", on_submit=process_message, key="prompt"
)



from dotenv import load_dotenv
from langchain_openai import ChatOpenAI

load_dotenv()

import chainlit as cl
from langchain.prompts import ChatPromptTemplate
from langchain.schema import StrOutputParser
from langchain.schema.runnable import Runnable
from langchain.schema.runnable.config import RunnableConfig

import os
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk import trace as trace_sdk
from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SimpleSpanProcessor

from opentelemetry.instrumentation.langchain import LangchainInstrumentor

# Set up OpenTelemetry trace provider with LangWatch as the endpoint
tracer_provider = trace_sdk.TracerProvider()
tracer_provider.add_span_processor(
    SimpleSpanProcessor(
        OTLPSpanExporter(
            endpoint=f"{os.environ.get('LANGWATCH_ENDPOINT', 'https://app.langwatch.ai')}/api/otel/v1/traces",
            headers={"Authorization": "Bearer " + os.environ["LANGWATCH_API_KEY"]},
        )
    )
)
# Optionally, you can also print the spans to the console.
tracer_provider.add_span_processor(SimpleSpanProcessor(ConsoleSpanExporter()))

LangchainInstrumentor().instrument(tracer_provider=tracer_provider)


@cl.on_chat_start
async def on_chat_start():
    model = ChatOpenAI(streaming=True)
    prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "You are a helpful assistant that only reply in short tweet-like responses, using lots of emojis.",
            ),
            ("human", "{question}"),
        ]
    )
    runnable = prompt | model | StrOutputParser()
    cl.user_session.set("runnable", runnable)


@cl.on_message
async def main(message: cl.Message):
    runnable: Runnable = cl.user_session.get("runnable")  # type: ignore

    msg = cl.Message(content="")

    async for chunk in runnable.astream(
        {"question": message.content},
        config=RunnableConfig(
            callbacks=[
                cl.LangchainCallbackHandler(),
            ],
            metadata={"user_id": "123", "thread_id": "789", "customer_id": "456"},
        ),
    ):
        await msg.stream_token(chunk)

    await msg.send()
